"""Generate Caddy configuration from the database."""
from __future__ import annotations

from pathlib import Path

from sqlalchemy import select

from . import models
from .db import session_scope
from .helper_runner import install_generated_file
from .config import ensure_cache_dir
from .importer import DEFAULT_CONFIG_NAME


class ExportError(RuntimeError):
    pass


def render_caddyfile_text(
    db_path: Path | None = None,
    *,
    snapshot_kind: models.SnapshotKind = models.SNAPSHOT_KIND_CADDY_TUI,
) -> str:
    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
        if config is None:
            return ""
        snapshot = session.scalar(
            select(models.ConfigSnapshot)
                .where(
                    models.ConfigSnapshot.config_id == config.id,
                    models.ConfigSnapshot.source_kind == snapshot_kind,
                )
                .limit(1)
        )
        if snapshot is None:
            return ""
        chunks: list[str] = []
        for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index):
            if block.raw_prelude:
                chunks.append(block.raw_prelude)
            fragments = sorted(block.fragments, key=lambda f: f.fragment_index)
            if fragments:
                chunks.extend(fragment.content for fragment in fragments)
            else:
                chunks.append(_synthesise_block(block))
            if block.raw_postlude:
                chunks.append(block.raw_postlude)
        return "".join(chunks)


def _synthesise_block(block: models.ServerBlock) -> str:
    labels = ", ".join(site.raw_label for site in sorted(block.sites, key=lambda s: s.label_index))
    header = f"{labels} {{\n" if labels else "{\n"
    body = "    respond \"caddy-tui placeholder\"\n"
    return header + body + "}\n"


def generate_caddyfile(
    target: Path,
    db_path: Path | None = None,
    *,
    snapshot_kind: models.SnapshotKind = models.SNAPSHOT_KIND_CADDY_TUI,
) -> Path:
    data = render_caddyfile_text(db_path=db_path, snapshot_kind=snapshot_kind)
    try:
        target.write_text(data)
    except PermissionError:
        cache = ensure_cache_dir() / "generated"
        cache.mkdir(parents=True, exist_ok=True)
        staged = cache / target.name
        staged.write_text(data)
        success, command, error = install_generated_file(staged, target)
        if not success:
            detail = error or "Helper install failed"
            hint = f"Run: {command}" if command else "Run helper install manually"
            raise PermissionError(f"Unable to write {target}: {detail}. {hint}")
    return target



"""Helpers for working with configuration snapshots."""
from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable
import json
import tempfile
from hashlib import sha256

from sqlalchemy import select
from sqlalchemy.orm import Session

from . import models
from .caddy_integration import CaddyError, adapt_caddyfile
from .json_normalizer import blocks_from_caddy_json
from .db import session_scope
from .importer import DEFAULT_CONFIG_NAME
from .live_renderer import render_live_block_like_caddyfile

SNAPSHOT_PAIRINGS: tuple[tuple[models.SnapshotKind, models.SnapshotKind], ...] = (
    (models.SNAPSHOT_KIND_CADDY_TUI, models.SNAPSHOT_KIND_CADDYFILE),
    (models.SNAPSHOT_KIND_CADDY_TUI, models.SNAPSHOT_KIND_CADDY_LIVE),
    (models.SNAPSHOT_KIND_CADDYFILE, models.SNAPSHOT_KIND_CADDY_LIVE),
)

SNAPSHOT_LABELS = {
    models.SNAPSHOT_KIND_CADDY_TUI: "caddy-tui",
    models.SNAPSHOT_KIND_CADDYFILE: "caddyfile",
    models.SNAPSHOT_KIND_CADDY_LIVE: "caddy live",
}
CADDYFILE_HIDE_SENTINEL = "__caddyfile__"


@dataclass(slots=True)
class SnapshotComparison:
    left_kind: models.SnapshotKind
    right_kind: models.SnapshotKind
    status: str
    mismatch_count: int | None
    left_hash: str | None
    right_hash: str | None


@dataclass(slots=True)
class SnapshotBlockText:
    block_index: int
    text: str
    key: tuple[str, ...]
    handles: tuple[str, ...]
    handlers: tuple[str, ...]
    hosts: tuple[str, ...]
    roots: tuple[str, ...]
    paths: tuple[str, ...]
    groups: tuple[str, ...]
    encodings: tuple[str, ...]
    locations: tuple[str, ...]
    dials: tuple[str, ...]
    status_codes: tuple[str, ...]
    route_payloads: tuple[str, ...]


def get_snapshot(session: Session, config_id: int, kind: models.SnapshotKind) -> models.ConfigSnapshot | None:
    return session.scalar(
        select(models.ConfigSnapshot)
        .where(
            models.ConfigSnapshot.config_id == config_id,
            models.ConfigSnapshot.source_kind == kind,
        )
        .limit(1)
    )


def structural_hash(snapshot: models.ConfigSnapshot) -> str:
    route_blobs = _snapshot_route_blobs(snapshot)
    if route_blobs is not None:
        blob = json.dumps(route_blobs, sort_keys=True, ensure_ascii=False)
        return sha256(blob.encode("utf-8")).hexdigest()
    payload = [_block_payload(block) for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index)]
    blob = json.dumps(payload, sort_keys=True, ensure_ascii=False)
    return sha256(blob.encode("utf-8")).hexdigest()


def compare_snapshots(
    left: models.ConfigSnapshot | None,
    right: models.ConfigSnapshot | None,
    *,
    left_kind: models.SnapshotKind,
    right_kind: models.SnapshotKind,
) -> SnapshotComparison:
    if left is None or right is None:
        return SnapshotComparison(
            left_kind=left_kind,
            right_kind=right_kind,
            status="missing",
            mismatch_count=None,
            left_hash=structural_hash(left) if left else None,
            right_hash=structural_hash(right) if right else None,
        )

    left_blocks = _block_hashes(left)
    right_blocks = _block_hashes(right)
    mismatch = 0
    for block_index in sorted(set(left_blocks) | set(right_blocks)):
        if left_blocks.get(block_index) != right_blocks.get(block_index):
            mismatch += 1

    left_hash = structural_hash(left)
    right_hash = structural_hash(right)
    status = "match" if left_hash == right_hash else "different"
    return SnapshotComparison(
        left_kind=left_kind,
        right_kind=right_kind,
        status=status,
        mismatch_count=mismatch,
        left_hash=left_hash,
        right_hash=right_hash,
    )


def _block_hashes(snapshot: models.ConfigSnapshot) -> dict[int, str]:
    route_blobs = _snapshot_route_blobs(snapshot)
    if route_blobs is not None:
        return {
            idx: sha256(blob.encode("utf-8")).hexdigest()
            for idx, blob in enumerate(route_blobs)
        }
    return {
        block.block_index: sha256(
            json.dumps(_block_payload(block), sort_keys=True, ensure_ascii=False).encode("utf-8")
        ).hexdigest()
        for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index)
    }


def _block_payload(block: models.ServerBlock) -> dict:
    return {
        "index": block.block_index,
        "is_global": block.is_global,
        "raw_prelude": block.raw_prelude,
        "raw_postlude": block.raw_postlude,
        "sites": [
            {
                "label": site.raw_label,
                "host": site.host,
                "port": site.port,
                "scheme": site.scheme,
                "is_ipv6": site.is_ipv6,
                "is_wildcard": site.is_wildcard,
                "order": site.label_index,
            }
            for site in sorted(block.sites, key=lambda site: site.label_index)
        ],
        "fragments": [
            {
                "kind": fragment.kind,
                "content": fragment.content,
                "index": fragment.fragment_index,
            }
            for fragment in sorted(block.fragments, key=lambda frag: frag.fragment_index)
        ],
        "directives": [
            {
                "name": directive.name,
                "matcher": directive.matcher,
                "line": directive.line_index,
                "raw_leading": directive.raw_leading,
                "raw_trailing": directive.raw_trailing,
                "has_block": directive.has_block,
                "raw_block_body": directive.raw_block_body,
                "args": [
                    {
                        "value": arg.value,
                        "index": arg.arg_index,
                    }
                    for arg in sorted(directive.args, key=lambda arg: arg.arg_index)
                ],
                "kv": [
                    {
                        "section": kv.section,
                        "key": kv.key,
                        "value": kv.value,
                        "index": kv.kv_index,
                    }
                    for kv in sorted(directive.kv_pairs, key=lambda kv: kv.kv_index)
                ],
            }
            for directive in sorted(block.directives, key=lambda d: d.line_index)
        ],
    }


def load_snapshot_block_texts(db_path: Path, kind: models.SnapshotKind) -> list[SnapshotBlockText]:
    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
        if not config:
            return []
        snapshot = get_snapshot(session, config.id, kind)
        if snapshot is None:
            return []
        scrub_paths = _snapshot_scrub_paths(snapshot)
        route_lookup: dict[tuple[str, ...], deque[str]] = {}
        if kind in {models.SNAPSHOT_KIND_CADDY_TUI, models.SNAPSHOT_KIND_CADDYFILE}:
            route_lookup = _caddyfile_route_lookup(snapshot)
        block_texts: list[SnapshotBlockText] = []
        for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index):
            metadata = _block_json_metadata(block)
            block_key = _canonical_block_key(block)
            route_payloads: tuple[str, ...] = ()
            if kind == models.SNAPSHOT_KIND_CADDY_LIVE:
                route_payloads = tuple(_block_route_fragments(block, scrub_paths=scrub_paths))
            else:
                queue = route_lookup.get(block_key)
                if queue:
                    payload = queue.popleft()
                    route_payloads = (payload,)
                else:
                    route_payloads = tuple(_block_route_fragments(block, scrub_paths=scrub_paths))
            block_texts.append(
                SnapshotBlockText(
                    block_index=block.block_index,
                    text=_render_block_text(block, source_kind=kind),
                    key=block_key,
                    handles=metadata.handles,
                    handlers=metadata.handlers,
                    hosts=metadata.hosts,
                    roots=metadata.roots,
                    paths=metadata.paths,
                    groups=metadata.groups,
                    encodings=metadata.encodings,
                    locations=metadata.locations,
                    dials=metadata.dials,
                    status_codes=metadata.status_codes,
                    route_payloads=route_payloads,
                )
            )
        return block_texts


def _render_block_text(block: models.ServerBlock, *, source_kind: models.SnapshotKind) -> str:
    if source_kind == models.SNAPSHOT_KIND_CADDY_LIVE:
        rendered = render_live_block_like_caddyfile(block)
        if rendered:
            trimmed = rendered.strip("\n")
            return trimmed if trimmed else rendered
    fragments = ''.join(fragment.content for fragment in sorted(block.fragments, key=lambda f: f.fragment_index))
    combined = f"{block.raw_prelude or ''}{fragments}{block.raw_postlude or ''}"
    stripped = combined.strip("\n")
    return stripped if stripped else combined


def _canonical_block_key(block: models.ServerBlock) -> tuple[str, ...]:
    labels = [site.raw_label.strip() for site in sorted(block.sites, key=lambda s: s.label_index) if site.raw_label]
    return _canonical_label_tuple(labels)


def _canonical_label_tuple(labels: Iterable[str], *, default: str = "(global)") -> tuple[str, ...]:
    cleaned = [label.strip() for label in labels if label and label.strip()]
    values = cleaned or [default]
    return tuple(sorted({value for value in values}))

def _snapshot_scrub_paths(snapshot: models.ConfigSnapshot) -> tuple[str, ...]:
    config = snapshot.config
    if config and getattr(config, "caddyfile_path", None):
        path = config.caddyfile_path.strip()
        return (path,) if path else ()
    return ()

def _snapshot_route_blobs(snapshot: models.ConfigSnapshot) -> list[str] | None:
    route_map = _snapshot_route_map(snapshot)
    if route_map is None:
        return None
    ordered: list[str] = []
    for key in sorted(route_map.keys(), key=_key_sort_value):
        ordered.extend(route_map[key])
    return ordered


def _routes_from_caddyfile_snapshot(snapshot: models.ConfigSnapshot) -> list[str]:
    return [payload for _key, payload in _caddyfile_route_entries(snapshot)]


def _routes_from_json_payload(payload: dict[str, Any], *, scrub_paths: tuple[str, ...]) -> list[str]:
    blobs: list[str] = []
    blocks = blocks_from_caddy_json(payload)
    for block in blocks:
        fragment = next((frag for frag in block.fragments if frag.kind == "json_route"), None)
        if fragment is None:
            continue
        normalised = _normalise_json_fragment(fragment.content, scrub_paths=scrub_paths)
        if normalised is None:
            continue
        blobs.append(normalised)
    return blobs


def render_snapshot_text(snapshot: models.ConfigSnapshot) -> str:
    chunks: list[str] = []
    for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index):
        if block.raw_prelude:
            chunks.append(block.raw_prelude)
        fragments = sorted(block.fragments, key=lambda f: f.fragment_index)
        chunks.extend(fragment.content for fragment in fragments)
        if block.raw_postlude:
            chunks.append(block.raw_postlude)
    return "".join(chunks)


def _block_route_fragments(block: models.ServerBlock, *, scrub_paths: tuple[str, ...]) -> list[str]:
    payloads: list[str] = []
    for fragment in sorted(block.fragments, key=lambda f: f.fragment_index):
        if fragment.kind != "json_route":
            continue
        normalised = _normalise_json_fragment(fragment.content, scrub_paths=scrub_paths)
        if normalised:
            payloads.append(normalised)
    return payloads


def _snapshot_route_map(snapshot: models.ConfigSnapshot) -> dict[tuple[str, ...], list[str]] | None:
    kind = snapshot.source_kind
    if kind == models.SNAPSHOT_KIND_CADDY_LIVE:
        return _live_route_map(snapshot)
    if kind in {models.SNAPSHOT_KIND_CADDY_TUI, models.SNAPSHOT_KIND_CADDYFILE}:
        try:
            return _caddyfile_route_map(snapshot)
        except CaddyError:
            return None
    return None


def _live_route_map(snapshot: models.ConfigSnapshot) -> dict[tuple[str, ...], list[str]] | None:
    mapping: dict[tuple[str, ...], list[str]] = {}
    scrub_paths = _snapshot_scrub_paths(snapshot)
    for block in sorted(snapshot.server_blocks, key=lambda b: b.block_index):
        payloads = _block_route_fragments(block, scrub_paths=scrub_paths)
        if not payloads:
            return None
        key = _canonical_block_key(block)
        mapping.setdefault(key, []).extend(payloads)
    return mapping


def _caddyfile_route_map(snapshot: models.ConfigSnapshot) -> dict[tuple[str, ...], list[str]]:
    mapping: dict[tuple[str, ...], list[str]] = {}
    for key, payload in _caddyfile_route_entries(snapshot):
        mapping.setdefault(key, []).append(payload)
    return mapping


def _key_sort_value(key: tuple[str, ...]) -> tuple[int, str]:
    label = ", ".join(key)
    return (len(key), label.lower())


def _caddyfile_route_lookup(snapshot: models.ConfigSnapshot) -> dict[tuple[str, ...], deque[str]]:
    lookup: dict[tuple[str, ...], deque[str]] = {}
    try:
        entries = _caddyfile_route_entries(snapshot)
    except CaddyError:
        return lookup
    for key, payload in entries:
        lookup.setdefault(key, deque()).append(payload)
    return lookup


def _caddyfile_route_entries(snapshot: models.ConfigSnapshot) -> list[tuple[tuple[str, ...], str]]:
    text = render_snapshot_text(snapshot)
    if not text.strip():
        return []
    with tempfile.NamedTemporaryFile("w", delete=False) as temp:
        temp.write(text)
        temp_path = Path(temp.name)
    try:
        adapted = adapt_caddyfile(temp_path)
    finally:
        temp_path.unlink(missing_ok=True)
    entries: list[tuple[tuple[str, ...], str]] = []
    base_paths = _snapshot_scrub_paths(snapshot)
    scrub_paths = (*base_paths, str(temp_path))
    blocks = blocks_from_caddy_json(adapted)
    for block in blocks:
        key = _canonical_label_tuple(block.labels)
        fragment = next((frag for frag in block.fragments if frag.kind == "json_route"), None)
        if fragment is None:
            continue
        normalised = _normalise_json_fragment(fragment.content, scrub_paths=scrub_paths)
        if normalised is None:
            continue
        entries.append((key, normalised))
    return entries


def _normalise_json_fragment(content: str, *, scrub_paths: tuple[str, ...]) -> str | None:
    try:
        data = json.loads(content)
    except json.JSONDecodeError:
        return None
    if scrub_paths:
        _scrub_file_server_hide(data, scrub_paths)
    return json.dumps(data, sort_keys=True, separators=(",", ":"))


def _scrub_file_server_hide(node: Any, scrub_paths: tuple[str, ...]) -> None:
    if not scrub_paths:
        return
    if isinstance(node, dict):
        handler = node.get("handler")
        if handler == "file_server":
            hide_value = node.get("hide")
            hide_entries = _normalize_string_values(hide_value)
            if hide_entries:
                replaced = [CADDYFILE_HIDE_SENTINEL if entry in scrub_paths else entry for entry in hide_entries]
                node["hide"] = replaced
        for value in node.values():
            _scrub_file_server_hide(value, scrub_paths)
    elif isinstance(node, list):
        for item in node:
            _scrub_file_server_hide(item, scrub_paths)


@dataclass(slots=True)
class _BlockJsonMetadata:
    handles: tuple[str, ...]
    handlers: tuple[str, ...]
    hosts: tuple[str, ...]
    roots: tuple[str, ...]
    paths: tuple[str, ...]
    groups: tuple[str, ...]
    encodings: tuple[str, ...]
    locations: tuple[str, ...]
    dials: tuple[str, ...]
    status_codes: tuple[str, ...]


def _block_json_metadata(block: models.ServerBlock) -> _BlockJsonMetadata:
    handles: list[str] = []
    handlers: list[str] = []
    hosts: list[str] = []
    roots: list[str] = []
    paths: list[str] = []
    groups: list[str] = []
    encodings: list[str] = []
    locations: list[str] = []
    dials: list[str] = []
    status_codes: list[str] = []
    fragments = sorted(block.fragments, key=lambda f: f.fragment_index)
    for fragment in fragments:
        if fragment.kind != "json_route":
            continue
        try:
            route = json.loads(fragment.content)
        except json.JSONDecodeError:
            continue
        if not isinstance(route, dict):
            continue
        _collect_route_metadata(
            route,
            handles,
            handlers,
            hosts,
            roots,
            paths,
            groups,
            encodings,
            locations,
            dials,
            status_codes,
            prefix=(),
        )
    return _BlockJsonMetadata(
        handles=tuple(handles),
        handlers=tuple(handlers),
        hosts=tuple(_dedupe_preserve_order(hosts)),
        roots=tuple(_dedupe_preserve_order(roots)),
        paths=tuple(_dedupe_preserve_order(paths)),
        groups=tuple(_dedupe_preserve_order(groups)),
        encodings=tuple(_dedupe_preserve_order(encodings)),
        locations=tuple(_dedupe_preserve_order(locations)),
        dials=tuple(_dedupe_preserve_order(dials)),
        status_codes=tuple(_dedupe_preserve_order(status_codes)),
    )


def _collect_route_metadata(
    node: dict[str, Any] | None,
    handles: list[str],
    handlers: list[str],
    hosts: list[str],
    roots: list[str],
    paths: list[str],
    groups: list[str],
    encodings: list[str],
    locations: list[str],
    dials: list[str],
    status_codes: list[str],
    *,
    prefix: tuple[str, ...],
) -> None:
    if not isinstance(node, dict):
        return
    _extend_unique(hosts, _hosts_from_matchers(node))
    _extend_unique(paths, _paths_from_matchers(node))
    _extend_unique(groups, _groups_from_matchers(node))
    entries = _normalise_handle_entries(node)
    for idx, entry in enumerate(entries):
        path_parts = (*prefix, f"handle[{idx}]")
        raw_handler = entry.get("handler") if isinstance(entry, dict) else None
        handler_name = _normalize_handler_name(raw_handler)
        label = ".".join(path_parts)
        if handler_name:
            _extend_unique(roots, _root_values(entry))
            _extend_unique(encodings, _encoding_values(entry))
            _extend_unique(locations, _location_values(entry))
            _extend_unique(dials, _dial_values(entry))
            _extend_unique(paths, _handler_path_values(entry))
            _extend_unique(status_codes, _status_code_values(entry))
            if handler_name != "subroute":
                handles.append(f"{label}: {handler_name}")
                if handler_name not in handlers:
                    handlers.append(handler_name)
        _recurse_nested_routes(
            entry,
            handles,
            handlers,
            hosts,
            roots,
            paths,
            groups,
            encodings,
            locations,
            dials,
            status_codes,
            parent_path=path_parts,
        )

    nested_routes = node.get("routes") if isinstance(node, dict) else None
    if isinstance(nested_routes, list):
        for idx, route in enumerate(nested_routes):
            if isinstance(route, dict):
                _collect_route_metadata(
                    route,
                    handles,
                    handlers,
                    hosts,
                    roots,
                    paths,
                    groups,
                    encodings,
                    locations,
                    dials,
                    status_codes,
                    prefix=(*prefix, f"routes[{idx}]"),
                )


def _normalise_handle_entries(node: dict[str, Any] | None) -> list[dict[str, Any]]:
    if not isinstance(node, dict):
        return []
    handle_value = node.get("handle")
    handlers_value = node.get("handlers")
    entries = handle_value or handlers_value
    if entries is None:
        return []
    if isinstance(entries, dict):
        return [entries]
    if isinstance(entries, list):
        return [entry for entry in entries if isinstance(entry, dict)]
    return []


def _recurse_nested_routes(
    entry: dict[str, Any] | None,
    handles: list[str],
    handlers: list[str],
    hosts: list[str],
    roots: list[str],
    paths: list[str],
    groups: list[str],
    encodings: list[str],
    locations: list[str],
    dials: list[str],
    status_codes: list[str],
    *,
    parent_path: tuple[str, ...],
) -> None:
    if not isinstance(entry, dict):
        return
    nested_routes = entry.get("routes")
    if not isinstance(nested_routes, list):
        return
    for idx, route in enumerate(nested_routes):
        if isinstance(route, dict):
            _collect_route_metadata(
                route,
                handles,
                handlers,
                hosts,
                roots,
                paths,
                groups,
                encodings,
                locations,
                dials,
                status_codes,
                prefix=(*parent_path, f"routes[{idx}]"),
            )


def _root_values(entry: dict[str, Any]) -> list[str]:
    value = entry.get("root")
    if isinstance(value, str):
        return [value] if value else []
    if isinstance(value, Iterable):
        roots: list[str] = []
        for item in value:
            if isinstance(item, str) and item:
                roots.append(item)
        return roots
    return []


def _hosts_from_matchers(node: dict[str, Any]) -> list[str]:
    matchers = node.get("match")
    if not isinstance(matchers, list):
        return []
    hosts: list[str] = []
    for matcher in matchers:
        if not isinstance(matcher, dict):
            continue
        hosts.extend(_normalize_host_values(matcher.get("host")))
        hosts.extend(_normalize_host_values(matcher.get("hosts")))
    return hosts


def _normalize_host_values(value: Any) -> list[str]:
    return _normalize_string_values(value)


def _paths_from_matchers(node: dict[str, Any]) -> list[str]:
    matchers = node.get("match")
    if not isinstance(matchers, list):
        return []
    paths: list[str] = []
    for matcher in matchers:
        if not isinstance(matcher, dict):
            continue
        paths.extend(_normalize_string_values(matcher.get("paths")))
        paths.extend(_normalize_string_values(matcher.get("path")))
    return paths


def _groups_from_matchers(node: dict[str, Any]) -> list[str]:
    matchers = node.get("match")
    if not isinstance(matchers, list):
        return []
    groups: list[str] = []
    for matcher in matchers:
        if not isinstance(matcher, dict):
            continue
        group = matcher.get("group")
        if isinstance(group, str) and group:
            groups.append(group)
    return groups


def _extend_unique(target: list[str], items: Iterable[str]) -> None:
    for item in items:
        if item and item not in target:
            target.append(item)


def _dedupe_preserve_order(items: Iterable[str]) -> list[str]:
    seen: set[str] = set()
    ordered: list[str] = []
    for item in items:
        if item in seen:
            continue
        seen.add(item)
        ordered.append(item)
    return ordered


def _normalize_string_values(value: Any) -> list[str]:
    if value is None:
        return []
    if isinstance(value, str):
        return [value] if value else []
    if isinstance(value, Iterable):
        results: list[str] = []
        for entry in value:
            if isinstance(entry, str) and entry:
                results.append(entry)
        return results
    return []


def _encoding_values(entry: dict[str, Any]) -> list[str]:
    value = entry.get("encodings")
    if isinstance(value, dict):
        return [str(key) for key in value.keys() if key]
    return _normalize_string_values(value)


def _location_values(entry: dict[str, Any]) -> list[str]:
    return _normalize_string_values(entry.get("location"))


def _dial_values(entry: dict[str, Any]) -> list[str]:
    values: list[str] = []
    values.extend(_normalize_string_values(entry.get("dial")))
    upstreams = entry.get("upstreams")
    if isinstance(upstreams, list):
        for upstream in upstreams:
            if isinstance(upstream, dict):
                dial = upstream.get("dial")
                if isinstance(dial, str) and dial:
                    values.append(dial)
    return values


def _handler_path_values(entry: dict[str, Any]) -> list[str]:
    values = _normalize_string_values(entry.get("path"))
    values.extend(_normalize_string_values(entry.get("paths")))
    return values


def _status_code_values(entry: dict[str, Any]) -> list[str]:
    value = entry.get("status_code")
    if value is None:
        return []
    if isinstance(value, int):
        return [str(value)]
    if isinstance(value, str):
        trimmed = value.strip()
        return [trimmed] if trimmed else []
    if isinstance(value, Iterable):
        codes: list[str] = []
        for item in value:
            if isinstance(item, int):
                codes.append(str(item))
            elif isinstance(item, str):
                trimmed = item.strip()
                if trimmed:
                    codes.append(trimmed)
        return codes
    return []


def _normalize_handler_name(name: str | None) -> str | None:
    if not name:
        return None
    lowered = name.strip().lower()
    if not lowered:
        return None
    if lowered in {"subroute", "log", "rewrite"}:
        return None
    if lowered == "headers":
        return "header"
    return lowered



from pathlib import Path

from caddy_tui import db
from caddy_tui.models import (
    Config,
    ConfigSnapshot,
    RawFragment,
    ServerBlock,
    ServerBlockSite,
    SNAPSHOT_KIND_CADDY_TUI,
)
from caddy_tui.exporter import generate_caddyfile, render_caddyfile_text


def test_generate_basic_caddyfile(tmp_path: Path):
    db_path = tmp_path / "config.db"
    db._engine = None  # type: ignore[attr-defined]
    db._SessionLocal = None  # type: ignore[attr-defined]
    db.init_db(db_path)
    sample_block = "example.com {\n    respond \"ok\"\n}\n"
    with db.session_scope(db_path) as session:
        config = Config(name="default", caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()
        snapshot = ConfigSnapshot(config=config, source_kind=SNAPSHOT_KIND_CADDY_TUI)
        session.add(snapshot)
        session.flush()
        block = ServerBlock(snapshot=snapshot, block_index=0, raw_prelude=None, raw_postlude=None)
        session.add(block)
        session.flush()
        block.sites.append(
            ServerBlockSite(
                raw_label="example.com",
                host="example.com",
                port=None,
                scheme=None,
                is_ipv6=False,
                is_wildcard=False,
                label_index=0,
            )
        )
        block.fragments.append(RawFragment(fragment_index=0, kind="block", content=sample_block))

    target = tmp_path / "Caddyfile.generated"
    generate_caddyfile(target, db_path=db_path)
    assert sample_block.strip() in target.read_text()
    rendered = render_caddyfile_text(db_path=db_path)
    assert sample_block.strip() in rendered



"""Integration helpers for invoking the caddy binary."""
from __future__ import annotations

import json
import subprocess
from pathlib import Path
from shutil import which
from typing import Any

from .config import AppPaths, CADDY_BIN


class CaddyError(RuntimeError):
    pass


def _caddy_bin(paths: AppPaths | None = None) -> str:
    configured = (paths.caddy_bin if paths else None) or CADDY_BIN
    candidate = configured or which("caddy")
    if not candidate:
        raise CaddyError("Unable to locate caddy binary. Set CADDY_TUI_CADDY_BIN.")
    return candidate


def adapt_caddyfile(path: Path, *, paths: AppPaths | None = None) -> dict[str, Any]:
    bin_path = _caddy_bin(paths)
    proc = subprocess.run(
        [bin_path, "adapt", "--config", str(path), "--adapter", "caddyfile", "--pretty"],
        check=False,
        capture_output=True,
        text=True,
    )
    if proc.returncode != 0:
        raise CaddyError(proc.stderr.strip() or "caddy adapt failed")
    return json.loads(proc.stdout)


def validate_config(config_path: Path, fmt: str = "caddyfile", *, paths: AppPaths | None = None) -> None:
    bin_path = _caddy_bin(paths)
    cmd = [bin_path, "validate", "--config", str(config_path)]
    if fmt == "caddyfile":
        cmd.extend(["--adapter", "caddyfile"])
    proc = subprocess.run(cmd, check=False, capture_output=True, text=True)
    if proc.returncode != 0:
        raise CaddyError(proc.stderr.strip() or "caddy validate failed")


def reload_caddy(config_path: Path, fmt: str = "caddyfile", *, paths: AppPaths | None = None) -> None:
    bin_path = _caddy_bin(paths)
    cmd = [bin_path, "reload", "--config", str(config_path)]
    if fmt == "caddyfile":
        cmd.extend(["--adapter", "caddyfile"])
    elif fmt == "json":
        cmd.extend(["--adapter", "json"])
    proc = subprocess.run(cmd, check=False, capture_output=True, text=True)
    if proc.returncode != 0:
        raise CaddyError(proc.stderr.strip() or "caddy reload failed")



"""Lightweight parser that segments a Caddyfile into server blocks."""
from __future__ import annotations

from dataclasses import dataclass
from typing import List


@dataclass(slots=True)
class ParsedFragment:
    kind: str
    content: str


@dataclass(slots=True)
class ParsedBlock:
    labels: list[str]
    is_global: bool
    raw_prelude: str
    raw_postlude: str
    fragments: list[ParsedFragment]


@dataclass(slots=True)
class ParsedConfig:
    blocks: list[ParsedBlock]


class CaddyfileParseError(RuntimeError):
    pass


def parse_caddyfile_text(text: str) -> ParsedConfig:
    """Parse the Caddyfile text into ordered blocks.

    This parser is intentionally conservative: it only cares about brace
    balancing so it can faithfully capture the text for each server block while
    also recording the host labels that appear in the header. Comments and
    whitespace between blocks are preserved via the per-block ``raw_prelude``
    and ``raw_postlude`` fields.
    """

    length = len(text)
    pos = 0
    blocks: list[ParsedBlock] = []
    pending_ws = ""

    pos, leading = _consume_ws_and_comments(text, pos)
    pending_ws += leading

    while pos < length:
        brace_index = _find_next_char(text, "{", pos)
        if brace_index == -1:
            # Remainder is trailing whitespace/comments.
            pending_ws += text[pos:]
            break
        header_text = text[pos:brace_index]
        labels = _split_labels(header_text)
        is_global = len(labels) == 0

        block_start = pos
        pos = brace_index + 1
        body_start = pos

        closing_index = _find_matching_brace(text, brace_index)
        if closing_index is None:
            raise CaddyfileParseError("Unbalanced braces in Caddyfile")
        block_end = closing_index + 1

        body_text = text[body_start:closing_index]
        header_segment = text[block_start:brace_index + 1]
        footer_segment = text[closing_index:block_end]

        block = ParsedBlock(
            labels=labels,
            is_global=is_global,
            raw_prelude=pending_ws,
            raw_postlude="",
            fragments=[
                ParsedFragment(kind="header", content=header_segment),
                ParsedFragment(kind="body", content=body_text),
                ParsedFragment(kind="footer", content=footer_segment),
            ],
        )
        blocks.append(block)
        pending_ws = ""
        pos = block_end
        pos, consumed = _consume_ws_and_comments(text, pos)
        pending_ws += consumed

    if blocks:
        blocks[-1].raw_postlude = pending_ws
    elif pending_ws:
        # Entire file is whitespace/comments; synthesise a block so we can
        # retain the content.
        blocks.append(
            ParsedBlock(
                labels=[],
                is_global=True,
                raw_prelude="",
                raw_postlude=pending_ws,
                fragments=[],
            )
        )

    return ParsedConfig(blocks=blocks)


def _consume_ws_and_comments(text: str, pos: int) -> tuple[int, str]:
    length = len(text)
    fragments: List[str] = []
    while pos < length:
        ch = text[pos]
        if ch in " \t\r\n":
            start = pos
            while pos < length and text[pos] in " \t\r\n":
                pos += 1
            fragments.append(text[start:pos])
        elif ch == "#":
            start = pos
            newline = text.find("\n", pos)
            if newline == -1:
                fragments.append(text[start:])
                pos = length
            else:
                fragments.append(text[start : newline + 1])
                pos = newline + 1
        else:
            break
    return pos, "".join(fragments)


def _find_next_char(text: str, target: str, start: int) -> int:
    idx = text.find(target, start)
    return idx if idx != -1 else -1


def _find_matching_brace(text: str, open_index: int) -> int | None:
    depth = 0
    idx = open_index
    length = len(text)
    while idx < length:
        ch = text[idx]
        if ch == "{":
            depth += 1
        elif ch == "}":
            depth -= 1
            if depth == 0:
                return idx
        idx += 1
    return None


def _split_labels(header_text: str) -> list[str]:
    cleaned = header_text.replace("\n", " ")
    tokens: list[str] = []
    for part in cleaned.split():
        for chunk in part.split(","):
            value = chunk.strip()
            if value:
                tokens.append(value)
    return tokens


from pathlib import Path

from caddy_tui import helper_runner


class DummyCommand(helper_runner.HelperCommand):
    def __init__(self):
        super().__init__(["sudo", "caddy-tui-helper", "noop"])


def test_stage_caddyfile_copy_builds_command(monkeypatch, tmp_path):
    source = tmp_path / "Caddyfile"
    source.write_text("test")

    monkeypatch.setattr(helper_runner, "_build_base_command", lambda **_kwargs: ["sudo", "helper"])
    monkeypatch.setattr(helper_runner, "_run_helper", lambda args: DummyCommand())
    staged, cmd, error = helper_runner.stage_caddyfile_copy(source)
    assert staged is not None
    assert cmd.endswith("noop")
    assert error is None


def test_install_generated_file_handles_failure(monkeypatch, tmp_path):
    source = tmp_path / "generated"
    source.write_text("data")

    def fail(args):
        raise helper_runner.HelperInvocationError(DummyCommand(), "denied")

    monkeypatch.setattr(helper_runner, "_build_base_command", lambda **_kwargs: ["sudo", "helper"])
    monkeypatch.setattr(helper_runner, "_run_helper", fail)
    success, cmd, error = helper_runner.install_generated_file(source, Path("/etc/caddy/Caddyfile"))
    assert not success
    assert cmd is not None
    assert error == "denied"


def test_check_caddy_service_success(monkeypatch):
    def fake_base_command(**_kwargs):
        return ["sudo", "helper"]

    def fake_run(args, capture_output=False):
        assert capture_output is True
        return DummyCommand(), "active"

    monkeypatch.setattr(helper_runner, "_build_base_command", fake_base_command)
    monkeypatch.setattr(helper_runner, "_run_helper", fake_run)
    state, cmd, error = helper_runner.check_caddy_service()
    assert state == "active"
    assert cmd is not None
    assert error is None


def test_check_caddy_service_failure(monkeypatch):
    def fake_base_command(**_kwargs):
        return ["sudo", "helper"]

    def fail(*args, **_kwargs):
        raise helper_runner.HelperInvocationError(DummyCommand(), "boom")

    monkeypatch.setattr(helper_runner, "_build_base_command", fake_base_command)
    monkeypatch.setattr(helper_runner, "_run_helper", fail)
    state, cmd, error = helper_runner.check_caddy_service()
    assert state is None
    assert cmd is not None
    assert error == "boom"


def test_restart_caddy_service(monkeypatch):
    calls: list[list[str]] = []

    def fake_base_command(**_kwargs):
        return ["sudo", "helper"]

    def fake_run(args, capture_output=False):
        calls.append(args)
        return DummyCommand()

    monkeypatch.setattr(helper_runner, "_build_base_command", fake_base_command)
    monkeypatch.setattr(helper_runner, "_run_helper", fake_run)
    success, cmd, error = helper_runner.restart_caddy_service()
    assert success is True
    assert cmd is not None
    assert error is None
    assert any(arg_list[-1] == "restart" for arg_list in calls)



from types import SimpleNamespace
from pathlib import Path

from caddy_tui import config


def test_determine_home_prefers_sudo_user(monkeypatch, tmp_path):
    monkeypatch.setenv("SUDO_USER", "dummy")
    monkeypatch.setattr(
        config,
        "pwd",
        SimpleNamespace(getpwnam=lambda user: SimpleNamespace(pw_dir=str(tmp_path))),
        raising=False,
    )
    assert config._determine_home() == tmp_path


def test_determine_home_defaults_to_path_home(monkeypatch):
    monkeypatch.delenv("SUDO_USER", raising=False)
    expected = Path.home()
    assert config._determine_home() == expected



"""Helpers for querying the Caddy admin API."""
from __future__ import annotations

from dataclasses import dataclass
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen
import json

from .caddyfile_parser import parse_caddyfile_text


@dataclass(slots=True)
class LiveApiStatus:
    state: str
    block_count: int | None
    caddyfile_text: str | None
    format: str
    json_payload: str | None = None
    error: str | None = None


def fetch_live_status(endpoint: str | None, *, timeout: float = 2.5) -> LiveApiStatus | None:
    if not endpoint:
        return None
    request = Request(endpoint)
    request.add_header("Accept", "text/caddyfile, text/plain, application/json")
    try:
        with urlopen(request, timeout=timeout) as response:
            content_type = response.headers.get("Content-Type", "")
            raw = response.read().decode("utf-8", errors="replace")
    except HTTPError as exc:  # pragma: no cover - depends on live service
        detail = exc.read().decode("utf-8", errors="replace")
        return LiveApiStatus(state="down", block_count=None, caddyfile_text=None, format="http", json_payload=None, error=detail or str(exc))
    except URLError as exc:  # pragma: no cover - depends on live service
        return LiveApiStatus(state="down", block_count=None, caddyfile_text=None, format="network", json_payload=None, error=str(exc))

    lowered = content_type.lower()
    if "caddyfile" in lowered or "text/plain" in lowered:
        return _from_caddyfile(raw)
    if "json" in lowered:
        return _from_json(raw)
    # Best effort: try to guess based on payload
    stripped = raw.lstrip()
    if stripped.startswith("{"):
        return _from_json(raw)
    return _from_caddyfile(raw)


def _from_caddyfile(text: str) -> LiveApiStatus:
    block_count: int | None
    try:
        parsed = parse_caddyfile_text(text)
        block_count = len(parsed.blocks)
    except Exception as exc:  # pragma: no cover - parser already tested elsewhere
        block_count = None
        text = text or ""
        return LiveApiStatus(state="live", block_count=block_count, caddyfile_text=text, format="caddyfile", json_payload=None, error=str(exc))
    return LiveApiStatus(state="live", block_count=block_count, caddyfile_text=text, format="caddyfile", json_payload=None)


def _from_json(payload: str) -> LiveApiStatus:
    try:
        data = json.loads(payload)
    except json.JSONDecodeError as exc:  # pragma: no cover - JSON parsing validated elsewhere
        return LiveApiStatus(state="live", block_count=None, caddyfile_text=None, format="json", json_payload=payload, error=str(exc))
    block_count = _count_http_routes(data)
    return LiveApiStatus(state="live", block_count=block_count, caddyfile_text=None, format="json", json_payload=payload)


def _count_http_routes(data: dict) -> int | None:
    apps = data.get("apps")
    if not isinstance(apps, dict):
        return None
    http_app = apps.get("http")
    if not isinstance(http_app, dict):
        return None
    servers = http_app.get("servers")
    if not isinstance(servers, dict):
        return None
    total = 0
    for server in servers.values():
        routes = server.get("routes") if isinstance(server, dict) else None
        if isinstance(routes, list):
            total += len(routes)
    return total



from caddy_tui import versioning


class _DummyResponse:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def read(self):
        return b'{"tag_name": "v0.2.0"}'


def test_fetch_latest_version_parses_tag(mocker):
    mocker.patch("urllib.request.urlopen", return_value=_DummyResponse())
    latest = versioning.fetch_latest_version(repo="ajphilton/caddy-tui")
    assert latest == "0.2.0"


def test_collect_version_info_handles_missing_remote(mocker):
    mocker.patch("caddy_tui.versioning.fetch_latest_version", return_value=None)
    info = versioning.collect_version_info()
    assert info.current == versioning.__version__
    assert info.latest is None
    assert info.update_available is False


"""Helpers that summarise application state for the TUI."""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from sqlalchemy import func, select
from sqlalchemy.exc import OperationalError

from .config import CADDY_ADMIN_ENDPOINT, CADDY_ADMIN_TIMEOUT, DB_PATH, LIVE_CADDYFILE
from .db import init_db, session_scope
from .importer import DEFAULT_CONFIG_NAME, import_caddyfile, import_caddyfile_text, import_caddy_json_payload
from .helper_runner import check_caddy_service
from .models import (
    Config,
    ConfigSnapshot,
    ServerBlock,
    ServerBlockSite,
    SnapshotKind,
    SNAPSHOT_KIND_CADDYFILE,
    SNAPSHOT_KIND_CADDY_LIVE,
    SNAPSHOT_KIND_CADDY_TUI,
)
from .live_api import LiveApiStatus, fetch_live_status
from .snapshots import SNAPSHOT_LABELS, SNAPSHOT_PAIRINGS, SnapshotComparison, compare_snapshots, get_snapshot


@dataclass(slots=True)
class SnapshotInfo:
    kind: SnapshotKind
    label: str
    available: bool
    site_count: int
    block_count: int
    collected_at: str | None
    source_path: str | None
    source_hash: str | None
    error: str | None = None


@dataclass(slots=True)
class AppStatus:
    db_path: Path
    db_exists: bool
    db_ready: bool
    last_import_path: str | None
    last_import_time: str | None
    snapshots: list[SnapshotInfo]
    comparisons: list[SnapshotComparison]
    block_count: int
    service_status: "ServiceStatus" | None


@dataclass(slots=True)
class ServiceStatus:
    state: str
    detail: str | None
    source: str | None
    block_count: int | None
    error: str | None = None


def collect_app_status(
    db_path: Path | None = None,
    *,
    refresh_sources: bool = True,
    refresh_live: bool = False,
    live_caddyfile: Path | None = None,
    purge_live_snapshot: bool = False,
    check_service: bool = True,
    admin_endpoint: str | None = CADDY_ADMIN_ENDPOINT,
) -> AppStatus:
    target = Path(db_path or DB_PATH)
    if not target.exists():
        try:
            init_db(db_path=target)
        except Exception:  # pragma: no cover - initial bootstrap best-effort
            pass
    db_exists = target.exists()
    db_ready = False
    last_import_path: str | None = None
    last_import_time: str | None = None
    snapshots: list[SnapshotInfo] = []
    comparisons: list[SnapshotComparison] = []
    block_count = 0
    refresh_errors: dict[SnapshotKind, str] = {}
    service_status: ServiceStatus | None = None
    live_api_status: LiveApiStatus | None = None
    if check_service:
        live_api_status = fetch_live_status(admin_endpoint, timeout=CADDY_ADMIN_TIMEOUT)

    if db_exists:
        try:
            config_info = _load_config_metadata(target)
            if refresh_sources and config_info is None:
                config_info = _bootstrap_caddyfile_from_disk(target, refresh_errors, live_caddyfile or LIVE_CADDYFILE)
            if refresh_sources and config_info:
                _refresh_snapshot_if_needed(
                    target,
                    config_info.caddyfile_path,
                    SNAPSHOT_KIND_CADDYFILE,
                    refresh_errors,
                )
            if refresh_live:
                _refresh_live_snapshot(
                    target,
                    live_api_status,
                    live_caddyfile or LIVE_CADDYFILE,
                    refresh_errors,
                    admin_endpoint,
                )
            snapshot_data = _collect_snapshot_details(target, purge_live_snapshot=purge_live_snapshot)
            if snapshot_data:
                db_ready = True
                last_import_path = snapshot_data.last_import_path
                last_import_time = snapshot_data.last_import_time
                snapshots = snapshot_data.snapshots
                comparisons = snapshot_data.comparisons
                block_count = snapshot_data.block_count
            elif config_info is None:
                db_ready = True
        except OperationalError:
            db_ready = False

    service_status = _build_service_status(live_api_status, live_caddyfile or LIVE_CADDYFILE, admin_endpoint)

    return AppStatus(
        db_path=target,
        db_exists=db_exists,
        db_ready=db_ready,
        last_import_path=last_import_path,
        last_import_time=last_import_time,
        snapshots=_apply_errors(snapshots, refresh_errors),
        comparisons=comparisons,
        block_count=block_count,
        service_status=service_status,
    )


@dataclass(slots=True)
class _SnapshotData:
    last_import_path: str | None
    last_import_time: str | None
    snapshots: list[SnapshotInfo]
    comparisons: list[SnapshotComparison]
    block_count: int


@dataclass(slots=True)
class _ConfigMetadata:
    caddyfile_path: Path | None


def _load_config_metadata(db_path: Path) -> _ConfigMetadata | None:
    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(Config).where(Config.name == DEFAULT_CONFIG_NAME))
        if not config:
            return None
        last_path = Path(config.caddyfile_path) if config.caddyfile_path else None
        return _ConfigMetadata(caddyfile_path=last_path)


def _bootstrap_caddyfile_from_disk(
    db_path: Path,
    errors: dict[SnapshotKind, str],
    fallback_path: Path | None,
) -> _ConfigMetadata | None:
    try:
        summary = import_caddyfile(fallback_path, target_snapshot=SNAPSHOT_KIND_CADDY_TUI, mirror_to=(SNAPSHOT_KIND_CADDYFILE,), db_path=db_path)
    except FileNotFoundError:
        errors[SNAPSHOT_KIND_CADDY_TUI] = "No Caddyfile found; run an import"
        return None
    except Exception as exc:  # pragma: no cover - unexpected import failure surfaces to UI
        errors[SNAPSHOT_KIND_CADDY_TUI] = str(exc)
        return None
    return _ConfigMetadata(caddyfile_path=Path(summary.source_path))


def _refresh_snapshot_if_needed(
    db_path: Path,
    path: Path | None,
    kind: SnapshotKind,
    errors: dict[SnapshotKind, str],
    *,
    required: bool = False,
) -> None:
    if not path:
        if required:
            errors[kind] = "Live Caddyfile path not configured"
        return
    try:
        import_caddyfile(path, target_snapshot=kind, mirror_to=(), db_path=db_path)
    except Exception as exc:  # pragma: no cover - surfaced to status output
        errors[kind] = str(exc)


def _refresh_live_snapshot(
    db_path: Path,
    api_status: LiveApiStatus | None,
    fallback_path: Path | None,
    errors: dict[SnapshotKind, str],
    admin_endpoint: str | None,
) -> None:
    if api_status and api_status.caddyfile_text:
        try:
            import_caddyfile_text(
                api_status.caddyfile_text,
                source_label=admin_endpoint or "caddy-admin",
                target_snapshot=SNAPSHOT_KIND_CADDY_LIVE,
                db_path=db_path,
            )
            return
        except Exception as exc:  # pragma: no cover - parse/DB errors bubble to UI
            errors[SNAPSHOT_KIND_CADDY_LIVE] = str(exc)
            return
    if api_status and getattr(api_status, "json_payload", None):
        try:
            import_caddy_json_payload(
                api_status.json_payload,
                source_label=admin_endpoint or "caddy-admin",
                target_snapshot=SNAPSHOT_KIND_CADDY_LIVE,
                db_path=db_path,
            )
            return
        except Exception as exc:  # pragma: no cover - parse/DB errors bubble to UI
            errors[SNAPSHOT_KIND_CADDY_LIVE] = str(exc)
            return
    _refresh_snapshot_if_needed(
        db_path,
        fallback_path,
        SNAPSHOT_KIND_CADDY_LIVE,
        errors,
        required=True,
    )


def _collect_snapshot_details(db_path: Path, *, purge_live_snapshot: bool) -> _SnapshotData | None:
    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(Config).where(Config.name == DEFAULT_CONFIG_NAME))
        if not config:
            return None

        snapshots: list[SnapshotInfo] = []
        total_blocks = 0
        live_snapshot_obj: ConfigSnapshot | None = None
        for kind, label in SNAPSHOT_LABELS.items():
            snapshot = get_snapshot(session, config.id, kind)
            site_count = _snapshot_site_count(session, snapshot) if snapshot else 0
            block_count = _snapshot_block_count(session, snapshot) if snapshot else 0
            snapshots.append(
                SnapshotInfo(
                    kind=kind,
                    label=label,
                    available=snapshot is not None,
                    site_count=site_count,
                    block_count=block_count,
                    collected_at=snapshot.collected_at.isoformat(timespec="seconds") if snapshot else None,
                    source_path=snapshot.source_path if snapshot else None,
                    source_hash=snapshot.source_hash if snapshot else None,
                )
            )
            if kind == SNAPSHOT_KIND_CADDY_TUI:
                total_blocks = block_count
            if kind == SNAPSHOT_KIND_CADDY_LIVE:
                live_snapshot_obj = snapshot

        comparisons: list[SnapshotComparison] = []
        for left, right in SNAPSHOT_PAIRINGS:
            left_snapshot = get_snapshot(session, config.id, left)
            right_snapshot = get_snapshot(session, config.id, right)
            comparisons.append(
                compare_snapshots(
                    left_snapshot,
                    right_snapshot,
                    left_kind=left,
                    right_kind=right,
                )
            )

        data = _SnapshotData(
            last_import_path=config.caddyfile_path,
            last_import_time=config.last_imported_at,
            snapshots=snapshots,
            comparisons=comparisons,
            block_count=total_blocks,
        )
        if purge_live_snapshot and live_snapshot_obj is not None:
            session.delete(live_snapshot_obj)
        return data


def _snapshot_site_count(session, snapshot: ConfigSnapshot | None) -> int:
    if snapshot is None:
        return 0
    return (
        session.scalar(
            select(func.count())
            .select_from(ServerBlockSite)
            .join(ServerBlock, ServerBlockSite.block)
            .where(ServerBlock.snapshot_id == snapshot.id)
        )
        or 0
    )


def _snapshot_block_count(session, snapshot: ConfigSnapshot | None) -> int:
    if snapshot is None:
        return 0
    return session.scalar(select(func.count()).select_from(ServerBlock).where(ServerBlock.snapshot_id == snapshot.id)) or 0


def _apply_errors(snapshots: list[SnapshotInfo], errors: dict[SnapshotKind, str]) -> list[SnapshotInfo]:
    if not errors:
        return snapshots
    updated: list[SnapshotInfo] = []
    for snapshot in snapshots:
        error = errors.get(snapshot.kind)
        if error:
            updated.append(
                SnapshotInfo(
                    kind=snapshot.kind,
                    label=snapshot.label,
                    available=snapshot.available,
                    site_count=snapshot.site_count,
                    block_count=snapshot.block_count,
                    collected_at=snapshot.collected_at,
                    source_path=snapshot.source_path,
                    source_hash=snapshot.source_hash,
                    error=error,
                )
            )
        else:
            updated.append(snapshot)
    return updated


def refresh_live_snapshot(
    db_path: Path | None = None,
    *,
    live_caddyfile: Path | None = None,
) -> AppStatus:
    """Refresh the live snapshot without touching other sources."""
    return collect_app_status(
        db_path=db_path,
        refresh_sources=False,
        refresh_live=True,
        live_caddyfile=live_caddyfile,
        purge_live_snapshot=False,
        check_service=True,
        admin_endpoint=CADDY_ADMIN_ENDPOINT,
    )


def _build_service_status(
    api_status: LiveApiStatus | None,
    fallback_path: Path | None,
    admin_endpoint: str | None,
) -> ServiceStatus | None:
    if api_status:
        return ServiceStatus(
            state=api_status.state,
            detail=api_status.format,
            source=admin_endpoint,
            block_count=api_status.block_count,
            error=api_status.error,
        )
    if fallback_path is None:
        return None
    raw_state, command, error = check_caddy_service()
    if raw_state is None:
        return ServiceStatus(state="unknown", detail=None, source=command, block_count=None, error=error)
    normalized = _normalize_service_state(raw_state)
    return ServiceStatus(state=normalized, detail=raw_state, source=command, block_count=None, error=error)


def _normalize_service_state(raw: str) -> str:
    lowered = raw.lower().strip()
    if lowered in {"active", "running", "live", "ok"}:
        return "live"
    if lowered in {"inactive", "failed", "dead", "stopped", "down"}:
        return "down"
    return "unknown"



"""Import existing Caddy config into the database."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from hashlib import sha256
from pathlib import Path
from typing import Any, Sequence
import json
import os

from sqlalchemy import select

from . import models
from .caddyfile_parser import ParsedBlock, parse_caddyfile_text
from .caddy_integration import adapt_caddyfile
from .db import session_scope
from .helper_runner import stage_caddyfile_copy
from .json_normalizer import blocks_from_caddy_json

DEFAULT_CADDYFILE_PATHS: tuple[Path, ...] = (
    Path("/etc/caddy/Caddyfile"),
    Path("/usr/local/etc/caddy/Caddyfile"),
    Path("/etc/Caddyfile"),
    Path("./Caddyfile"),
)

MAX_PARENT_SEARCH_DEPTH = 5


def _generate_candidate_paths(explicit: Path) -> list[Path]:
    """Return a list of nearby paths that might contain a Caddyfile."""
    path = explicit.expanduser()
    candidates: list[Path] = []
    seen: set[Path] = set()

    def add(candidate: Path) -> None:
        candidate = candidate.expanduser()
        if candidate not in seen:
            seen.add(candidate)
            candidates.append(candidate)

    add(path)

    if path.name and path.name.lower() != "caddyfile":
        add(path.with_name("Caddyfile"))
    add(path / "Caddyfile")

    current = path.parent
    depth = 0
    while depth < MAX_PARENT_SEARCH_DEPTH and current != current.parent:
        add(current / "Caddyfile")
        current = current.parent
        depth += 1

    return candidates


def _resolve_explicit_path(explicit: Path) -> Path | None:
    for candidate in _generate_candidate_paths(explicit):
        if candidate.exists() and candidate.is_file():
            return candidate
    return None


class CaddyfilePermissionError(PermissionError):
    """Raised when the Caddyfile cannot be read due to permissions."""

    def __init__(self, path: Path, helper_command: str | None = None):
        self.path = path
        self.helper_command = helper_command
        message = f"Permission denied reading {path}. Run the import with elevated permissions or copy the file to a readable location."
        if helper_command:
            message = f"{message} You can run: {helper_command}"
        super().__init__(message)

    @property
    def suggested_command(self) -> str:
        if self.helper_command:
            return self.helper_command
        return f"sudo caddy-tui import --caddyfile {self.path}"


def find_caddyfile(explicit: Path | None = None) -> Path:
    if explicit:
        resolved = _resolve_explicit_path(explicit)
        if resolved:
            return resolved
    for candidate in DEFAULT_CADDYFILE_PATHS:
        if candidate.exists():
            return candidate
    raise FileNotFoundError("Unable to locate a Caddyfile to import")


@dataclass(slots=True)
class ImportSummary:
    source_path: Path
    site_labels: list[str]
    site_count: int
    snapshot_kind: models.SnapshotKind
    mirrored_snapshots: tuple[models.SnapshotKind, ...]


DEFAULT_CONFIG_NAME = "default"


def import_caddyfile(
    path: Path | None = None,
    *,
    helper_interactive: bool = False,
    target_snapshot: models.SnapshotKind = models.SNAPSHOT_KIND_CADDY_TUI,
    mirror_to: Sequence[models.SnapshotKind] | None = None,
    db_path: Path | None = None,
) -> ImportSummary:
    source = find_caddyfile(path)
    adapted_source = _ensure_accessible_source(source, helper_interactive=helper_interactive)
    adapt_caddyfile(adapted_source)  # validation only

    text = adapted_source.read_text()
    parsed = parse_caddyfile_text(text)
    if not parsed.blocks:
        raise ValueError("No server blocks detected in Caddyfile")

    digest = sha256(text.encode("utf-8")).hexdigest()
    collected_at = datetime.now(timezone.utc)
    timestamp = collected_at.isoformat(timespec="seconds")

    labels: list[str] = _summarise_block_labels(parsed.blocks)
    snapshots_written = _unique_kinds(target_snapshot, mirror_to)

    with session_scope(db_path=db_path) as session:
        config = _get_or_create_config(session, source)
        if target_snapshot == models.SNAPSHOT_KIND_CADDY_TUI:
            config.last_imported_at = timestamp
            config.last_caddyfile_hash = digest
        for kind in snapshots_written:
            _write_snapshot(
                session,
                config,
                kind,
                parsed.blocks,
                source_path=str(source),
                source_hash=digest,
                collected_at=collected_at,
            )

    return ImportSummary(
        source_path=source,
        site_labels=labels,
        site_count=len(labels),
        snapshot_kind=target_snapshot,
        mirrored_snapshots=tuple(kind for kind in snapshots_written if kind != target_snapshot),
    )


def import_caddyfile_text(
    text: str,
    *,
    source_label: str,
    target_snapshot: models.SnapshotKind,
    mirror_to: Sequence[models.SnapshotKind] | None = None,
    db_path: Path | None = None,
    require_config: bool = False,
) -> ImportSummary | None:
    parsed = parse_caddyfile_text(text)
    if not parsed.blocks:
        raise ValueError("No server blocks detected in Caddyfile text")

    digest = sha256(text.encode("utf-8")).hexdigest()
    collected_at = datetime.now(timezone.utc)
    timestamp = collected_at.isoformat(timespec="seconds")
    labels: list[str] = _summarise_block_labels(parsed.blocks)
    snapshots_written = _unique_kinds(target_snapshot, mirror_to)

    with session_scope(db_path=db_path) as session:
        config = _ensure_config_record(session, require_config=require_config, default_path=source_label)
        for kind in snapshots_written:
            _write_snapshot(
                session,
                config,
                kind,
                parsed.blocks,
                source_path=source_label,
                source_hash=digest,
                collected_at=collected_at,
            )

    return ImportSummary(
        source_path=Path(source_label),
        site_labels=labels,
        site_count=len(labels),
        snapshot_kind=target_snapshot,
        mirrored_snapshots=tuple(kind for kind in snapshots_written if kind != target_snapshot),
    )


def import_caddy_json_payload(
    payload: str | dict[str, Any],
    *,
    source_label: str,
    target_snapshot: models.SnapshotKind = models.SNAPSHOT_KIND_CADDY_LIVE,
    mirror_to: Sequence[models.SnapshotKind] | None = None,
    db_path: Path | None = None,
    require_config: bool = False,
) -> ImportSummary | None:
    data = json.loads(payload) if isinstance(payload, str) else payload
    blocks = blocks_from_caddy_json(data)
    normalised_text = json.dumps(data, sort_keys=True)
    digest = sha256(normalised_text.encode("utf-8")).hexdigest()
    collected_at = datetime.now(timezone.utc)
    labels: list[str] = _summarise_block_labels(blocks)
    snapshots_written = _unique_kinds(target_snapshot, mirror_to)

    with session_scope(db_path=db_path) as session:
        config = _ensure_config_record(session, require_config=require_config, default_path=source_label)
        for kind in snapshots_written:
            _write_snapshot(
                session,
                config,
                kind,
                blocks,
                source_path=source_label,
                source_hash=digest,
                collected_at=collected_at,
            )

    return ImportSummary(
        source_path=Path(source_label),
        site_labels=labels,
        site_count=len(labels),
        snapshot_kind=target_snapshot,
        mirrored_snapshots=tuple(kind for kind in snapshots_written if kind != target_snapshot),
    )


def _ensure_accessible_source(source: Path, *, helper_interactive: bool = False) -> Path:
    if os.access(source, os.R_OK):
        return source
    staged, command, _ = stage_caddyfile_copy(source, interactive=helper_interactive)
    if staged:
        return staged
    raise CaddyfilePermissionError(source, helper_command=command)


def _get_or_create_config(session, source: Path) -> models.Config:
    config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
    if config is None:
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path=str(source))
        session.add(config)
        session.flush()
    else:
        config.caddyfile_path = str(source)
    return config


def _ensure_config_record(session, *, require_config: bool, default_path: str | None) -> models.Config:
    config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
    if config is None:
        if require_config:
            raise RuntimeError("Initialise the database with caddy-tui init before importing text")
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path=default_path)
        session.add(config)
        session.flush()
    return config


def _write_snapshot(
    session,
    config: models.Config,
    kind: models.SnapshotKind,
    blocks: list[ParsedBlock],
    *,
    source_path: str | None,
    source_hash: str | None,
    collected_at: datetime,
) -> None:
    snapshot = session.scalar(
        select(models.ConfigSnapshot).where(
            models.ConfigSnapshot.config_id == config.id,
            models.ConfigSnapshot.source_kind == kind,
        )
    )
    if snapshot is None:
        snapshot = models.ConfigSnapshot(config=config, source_kind=kind)
        session.add(snapshot)
        session.flush()
    snapshot.source_path = source_path
    snapshot.source_hash = source_hash
    snapshot.collected_at = collected_at
    snapshot.server_blocks.clear()

    for index, block in enumerate(blocks):
        _store_block(snapshot, index, block)


def _store_block(snapshot: models.ConfigSnapshot, index: int, block: ParsedBlock) -> models.ServerBlock:
    block_model = models.ServerBlock(
        snapshot=snapshot,
        block_index=index,
        is_global=block.is_global,
        raw_prelude=block.raw_prelude or None,
        raw_postlude=block.raw_postlude or None,
    )
    snapshot.server_blocks.append(block_model)

    for label_index, raw_label in enumerate(block.labels):
        host, port, scheme, is_ipv6, is_wildcard = _analyse_label(raw_label)
        block_model.sites.append(
            models.ServerBlockSite(
                raw_label=raw_label,
                host=host,
                port=port,
                scheme=scheme,
                is_ipv6=is_ipv6,
                is_wildcard=is_wildcard,
                label_index=label_index,
            )
        )

    for fragment_index, fragment in enumerate(block.fragments):
        block_model.fragments.append(
            models.RawFragment(
                fragment_index=fragment_index,
                kind=fragment.kind,
                content=fragment.content,
            )
        )

    return block_model


def _summarise_block_labels(blocks: list[ParsedBlock]) -> list[str]:
    labels: list[str] = []
    for block in blocks:
        if block.labels:
            labels.append(", ".join(block.labels))
        else:
            labels.append("(global options)")
    return labels


def _unique_kinds(
    target_snapshot: models.SnapshotKind,
    mirror_to: Sequence[models.SnapshotKind] | None,
) -> tuple[models.SnapshotKind, ...]:
    if mirror_to is None and target_snapshot == models.SNAPSHOT_KIND_CADDY_TUI:
        requested = [models.SNAPSHOT_KIND_CADDY_TUI, models.SNAPSHOT_KIND_CADDYFILE]
    elif mirror_to is None:
        requested = [target_snapshot]
    else:
        requested = [target_snapshot, *mirror_to]
    seen: list[models.SnapshotKind] = []
    for kind in requested:
        if kind not in seen:
            seen.append(kind)
    return tuple(seen)


def _analyse_label(raw: str) -> tuple[str | None, int | None, str | None, bool, bool]:
    scheme: str | None = None
    host_port = raw
    if "://" in raw:
        scheme, host_port = raw.split("://", 1)

    host: str | None = None
    port: int | None = None
    is_ipv6 = False
    if host_port.startswith("["):
        end = host_port.find("]")
        if end != -1:
            host = host_port[1:end]
            remainder = host_port[end + 1 :]
            if remainder.startswith(":"):
                try:
                    port = int(remainder[1:])
                except ValueError:
                    port = None
            is_ipv6 = True
        else:
            host = host_port
    else:
        if ":" in host_port:
            maybe_host, maybe_port = host_port.rsplit(":", 1)
            if maybe_port.isdigit():
                host = maybe_host or None
                port = int(maybe_port)
            else:
                host = host_port
        else:
            host = host_port or None

    if host is None and port is None and host_port.startswith(":") and host_port[1:].isdigit():
        port = int(host_port[1:])

    is_wildcard = bool(host and "*" in host)
    return host, port, scheme, is_ipv6, is_wildcard



import json
from pathlib import Path

from sqlalchemy import select

from caddy_tui import db, models
from caddy_tui.importer import DEFAULT_CONFIG_NAME, import_caddy_json_payload, import_caddyfile_text
from caddy_tui.snapshots import compare_snapshots, load_snapshot_block_texts


def _reset_db(tmp_path: Path) -> Path:
    db._engine = None  # type: ignore[attr-defined]
    db._SessionLocal = None  # type: ignore[attr-defined]
    db_path = tmp_path / "config.db"
    db.init_db(db_path)
    return db_path


def _seed_snapshot(db_path: Path, kind: str) -> None:
    with db.session_scope(db_path=db_path) as session:
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()

        snapshot = models.ConfigSnapshot(config=config, source_kind=kind)
        session.add(snapshot)
        session.flush()

        block = models.ServerBlock(
            snapshot=snapshot,
            block_index=0,
            raw_prelude="",
            raw_postlude="\n",
        )
        session.add(block)
        session.flush()

        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=0,
                kind="header",
                content="example.com {\n",
            )
        )
        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=1,
                kind="body",
                content="    respond \"ok\"\n",
            )
        )
        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=2,
                kind="footer",
                content="}\n",
            )
        )
        block.sites.append(
            models.ServerBlockSite(
                block=block,
                raw_label="example.com",
                host="example.com",
                port=None,
                scheme=None,
                is_ipv6=False,
                is_wildcard=False,
                label_index=0,
            )
        )


def test_load_snapshot_block_texts_returns_content(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    _seed_snapshot(db_path, models.SNAPSHOT_KIND_CADDY_TUI)
    blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_TUI)
    assert len(blocks) == 1
    assert "example.com" in blocks[0].text
    assert "respond" in blocks[0].text
    assert blocks[0].key == ("example.com",)
    assert blocks[0].handles == ()
    assert blocks[0].handlers == ()
    assert blocks[0].hosts == ()
    assert blocks[0].roots == ()
    assert blocks[0].paths == ()
    assert blocks[0].groups == ()
    assert blocks[0].encodings == ()
    assert blocks[0].locations == ()
    assert blocks[0].dials == ()
    assert blocks[0].status_codes == ()
    assert len(blocks[0].route_payloads) == 1


def test_load_snapshot_block_texts_handles_missing_snapshot(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_LIVE)
    assert blocks == []


def test_load_snapshot_block_texts_global_block(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    with db.session_scope(db_path=db_path) as session:
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()
        snapshot = models.ConfigSnapshot(config=config, source_kind=models.SNAPSHOT_KIND_CADDY_TUI)
        session.add(snapshot)
        session.flush()
        block = models.ServerBlock(snapshot=snapshot, block_index=0, raw_prelude="", raw_postlude="")
        session.add(block)
        session.flush()
        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=0,
                kind="body",
                content="{\n    respond \"ok\"\n}\n",
            )
        )

    blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_TUI)
    assert blocks[0].key == ("(global)",)


def test_load_snapshot_block_texts_extracts_handle_metadata(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    with db.session_scope(db_path=db_path) as session:
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()

        snapshot = models.ConfigSnapshot(config=config, source_kind=models.SNAPSHOT_KIND_CADDY_TUI)
        session.add(snapshot)
        session.flush()

        block = models.ServerBlock(snapshot=snapshot, block_index=0, raw_prelude="", raw_postlude="")
        session.add(block)
        session.flush()

        route_payload = {
            "match": [
                {
                    "hosts": ["alpha.test", "beta.test"],
                    "paths": ["/api/*"],
                    "path": "/console",
                    "group": "medusa",
                }
            ],
            "handle": [
                {
                    "handler": "reverse_proxy",
                    "upstreams": [{"dial": "medusa:9000"}],
                },
                {
                    "handler": "encode",
                    "encodings": {"zstd": {}, "gzip": {}},
                },
                {
                    "handler": "subroute",
                    "routes": [
                        {
                            "handle": [
                                {"handler": "static_response", "status_code": 201},
                                {"handler": "rewrite", "location": "/tmp"},
                                    {"handler": "headers", "response": {"set": {"X-Test": ["1"]}}},
                                {},
                            ]
                        }
                    ],
                },
                {
                    "handler": "file_server",
                    "root": "/srv/site",
                    "location": "/shared/static",
                    "path": "/assets/*",
                },
            ],
        }

        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=0,
                kind="json_route",
                content=json.dumps(route_payload),
            )
        )

    blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_TUI)
    assert blocks[0].handles == (
        "handle[0]: reverse_proxy",
        "handle[1]: encode",
        "handle[2].routes[0].handle[0]: static_response",
        "handle[2].routes[0].handle[2]: header",
        "handle[3]: file_server",
    )
    assert blocks[0].handlers == (
        "reverse_proxy",
        "encode",
        "static_response",
        "header",
        "file_server",
    )
    assert blocks[0].hosts == ("alpha.test", "beta.test")
    assert blocks[0].roots == ("/srv/site",)
    assert blocks[0].paths == ("/api/*", "/console", "/assets/*")
    assert blocks[0].groups == ("medusa",)
    assert blocks[0].encodings == ("zstd", "gzip")
    assert blocks[0].locations == ("/shared/static",)
    assert blocks[0].dials == ("medusa:9000",)
    assert blocks[0].status_codes == ("201",)
    assert len(blocks[0].route_payloads) == 1


def test_live_snapshot_blocks_render_json_like_caddyfile(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    with db.session_scope(db_path=db_path) as session:
        config = models.Config(name=DEFAULT_CONFIG_NAME, caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()

        snapshot = models.ConfigSnapshot(config=config, source_kind=models.SNAPSHOT_KIND_CADDY_LIVE)
        session.add(snapshot)
        session.flush()

        block = models.ServerBlock(
            snapshot=snapshot,
            block_index=0,
            raw_prelude="# server: srv0 route: 0\n",
            raw_postlude="",
        )
        session.add(block)
        session.flush()

        block.sites.append(
            models.ServerBlockSite(
                block=block,
                raw_label="www.redirect.test",
                host="www.redirect.test",
                port=None,
                scheme=None,
                is_ipv6=False,
                is_wildcard=False,
                label_index=0,
            )
        )

        route_payload = {
            "match": [{"host": ["www.redirect.test"]}],
            "handle": [
                {
                    "handler": "subroute",
                    "routes": [
                        {
                            "handle": [
                                {
                                    "handler": "static_response",
                                    "headers": {"Location": ["https://redirect.test{http.request.uri}"]},
                                    "status_code": 308,
                                }
                            ]
                        }
                    ],
                }
            ],
            "terminal": True,
        }

        block.fragments.append(
            models.RawFragment(
                block=block,
                fragment_index=0,
                kind="json_route",
                content=json.dumps(route_payload, sort_keys=True),
            )
        )

    blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_LIVE)
    assert len(blocks) == 1
    text = blocks[0].text
    assert "www.redirect.test {" in text
    assert "redir https://redirect.test{http.request.uri} 308" in text
    assert text.strip().endswith("}")
    assert len(blocks[0].route_payloads) == 1


def test_compare_snapshots_uses_adapted_json(monkeypatch, tmp_path: Path):
    db_path = _reset_db(tmp_path)

    route_payload = {
        "match": [{"host": ["example.test"]}],
        "handle": [
            {
                "handler": "static_response",
                "body": "ok",
            }
        ],
        "terminal": True,
    }
    json_payload = {
        "apps": {
            "http": {
                "servers": {
                    "srv0": {
                        "routes": [route_payload],
                    }
                }
            }
        }
    }

    def fake_adapt(path: Path):  # type: ignore[unused-argument]
        return json_payload

    monkeypatch.setattr("caddy_tui.importer.adapt_caddyfile", fake_adapt)
    monkeypatch.setattr("caddy_tui.snapshots.adapt_caddyfile", fake_adapt)

    import_caddyfile_text(
        "example.test {\n    respond \"ok\"\n}\n",
        source_label="fs",
        target_snapshot=models.SNAPSHOT_KIND_CADDYFILE,
        db_path=db_path,
    )
    import_caddy_json_payload(
        json_payload,
        source_label="live",
        target_snapshot=models.SNAPSHOT_KIND_CADDY_LIVE,
        db_path=db_path,
    )

    with db.session_scope(db_path=db_path) as session:
        config = session.scalar(select(models.Config))
        assert config is not None
        file_snapshot = session.scalar(
            select(models.ConfigSnapshot)
            .where(
                models.ConfigSnapshot.config_id == config.id,
                models.ConfigSnapshot.source_kind == models.SNAPSHOT_KIND_CADDYFILE,
            )
            .limit(1)
        )
        live_snapshot = session.scalar(
            select(models.ConfigSnapshot)
            .where(
                models.ConfigSnapshot.config_id == config.id,
                models.ConfigSnapshot.source_kind == models.SNAPSHOT_KIND_CADDY_LIVE,
            )
            .limit(1)
        )
        assert file_snapshot is not None
        assert live_snapshot is not None
        comparison = compare_snapshots(
            file_snapshot,
            live_snapshot,
            left_kind=models.SNAPSHOT_KIND_CADDYFILE,
            right_kind=models.SNAPSHOT_KIND_CADDY_LIVE,
        )
        assert comparison.status == "match"
        assert comparison.mismatch_count == 0


def test_block_texts_include_route_payloads(monkeypatch, tmp_path: Path):
    db_path = _reset_db(tmp_path)

    route_payload = {
        "match": [{"host": ["payload.test"]}],
        "handle": [
            {
                "handler": "static_response",
                "body": "ok",
            }
        ],
        "terminal": True,
    }
    json_payload = {
        "apps": {
            "http": {
                "servers": {
                    "srv0": {
                        "routes": [route_payload],
                    }
                }
            }
        }
    }

    def fake_adapt(path: Path):  # type: ignore[unused-argument]
        return json_payload

    monkeypatch.setattr("caddy_tui.importer.adapt_caddyfile", fake_adapt)
    monkeypatch.setattr("caddy_tui.snapshots.adapt_caddyfile", fake_adapt)

    import_caddyfile_text(
        "payload.test {\n    respond \"ok\"\n}\n",
        source_label="fs",
        target_snapshot=models.SNAPSHOT_KIND_CADDYFILE,
        db_path=db_path,
    )
    import_caddy_json_payload(
        json_payload,
        source_label="live",
        target_snapshot=models.SNAPSHOT_KIND_CADDY_LIVE,
        db_path=db_path,
    )

    caddyfile_blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDYFILE)
    live_blocks = load_snapshot_block_texts(db_path, models.SNAPSHOT_KIND_CADDY_LIVE)

    assert len(caddyfile_blocks) == 1
    assert len(live_blocks) == 1
    assert caddyfile_blocks[0].route_payloads
    assert live_blocks[0].route_payloads
    assert caddyfile_blocks[0].route_payloads == live_blocks[0].route_payloads



"""Simple interactive terminal menu built with colorama and rich."""

from __future__ import annotations

import os
import shlex
import shutil
import subprocess
import tempfile
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Iterable

from colorama import Fore, Style, init as colorama_init
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .block_editor import blocks_to_text, load_caddy_tui_blocks, parse_single_block, save_caddy_tui_blocks
from .caddyfile_parser import ParsedBlock
from .config import LIVE_CADDYFILE
from .drift import compare_caddyfile, summarise_drift
from .exporter import generate_caddyfile
from .helper_runner import reload_caddy_service, restart_caddy_service
from .importer import CaddyfilePermissionError, import_caddyfile
from .models import SNAPSHOT_KIND_CADDYFILE, SNAPSHOT_KIND_CADDY_LIVE, SNAPSHOT_KIND_CADDY_TUI
from .snapshots import SNAPSHOT_LABELS, SnapshotBlockText, SnapshotComparison, load_snapshot_block_texts
from .status import AppStatus, ServiceStatus, SnapshotInfo, collect_app_status, refresh_live_snapshot
from .versioning import VersionInfo, collect_version_info


@dataclass(slots=True)
class MenuOption:
    key: str
    description: str
    handler: Callable[[], None]


class TerminalMenuApp:
    """Lightweight scrolling menu that prints status, results, and prompts."""

    def __init__(self) -> None:
        colorama_init(autoreset=True)
        self.console = Console()
        self.last_output: RenderableType | None = None
        self._last_output_once = False
        self.running = True
        self._refresh_live_next = True
        self._latest_status: AppStatus | None = None
        self._version_info: VersionInfo | None = None
        self._version_last_attempt: float | None = None

    def run(self) -> None:
        """Enter the interactive loop."""
        try:
            while self.running:
                status = collect_app_status(
                    live_caddyfile=LIVE_CADDYFILE,
                    refresh_live=self._refresh_live_next,
                )
                self._refresh_live_next = False
                self._latest_status = status
                self._get_version_info()
                options = self._build_menu(status)
                self._render_cycle(status, options)
                choice = input(f"{Fore.CYAN}Select option{Style.RESET_ALL}: ").strip().lower()
                if not choice:
                    continue
                option = next((opt for opt in options if opt.key == choice), None)
                if option is None:
                    self._set_message(f"Unknown option '{choice}'.", style="red")
                    continue
                try:
                    option.handler()
                except KeyboardInterrupt:
                    self._set_message("Action cancelled.", style="yellow")
        except KeyboardInterrupt:
            self.console.print("\nExiting...")

    # Rendering helpers

    def _render_cycle(self, status: AppStatus, options: list[MenuOption]) -> None:
        self.console.rule("caddy-tui")
        if self.last_output:
            self.console.print(self.last_output)
            self.console.print()
            if self._last_output_once:
                self.last_output = None
                self._last_output_once = False
        self.console.print(self._status_table(status))
        self.console.print()
        self.console.print("Menu:")
        for option in options:
            self.console.print(f"  [cyan]{option.key}[/cyan] — {option.description}")

    def _status_table(self, status: AppStatus) -> Table:
        table = Table(title="Status", show_lines=True)
        table.add_column("Field", style="bold")
        table.add_column("Value")
        table.add_row("DB path", str(status.db_path))
        table.add_row("DB ready", "yes" if status.db_ready else "no")
        table.add_row("Version", self._version_status_text())
        table.add_row("Stored blocks", str(status.block_count))
        if status.last_import_path:
            when = status.last_import_time or "time unknown"
            table.add_row("Last import", f"{status.last_import_path} ({when})")
        else:
            table.add_row("Last import", "Never — run an import")
        table.add_row("Snapshots", self._snapshot_matrix(status))
        for comparison in status.comparisons:
            label = f"{SNAPSHOT_LABELS[comparison.left_kind]} ↔ {SNAPSHOT_LABELS[comparison.right_kind]}"
            table.add_row(
                f"Diff ({label})",
                self._format_comparison(comparison),
            )
        return table

    def _snapshot_matrix(self, status: AppStatus) -> Table:
        matrix = Table(show_header=False, box=None, pad_edge=False)
        for _ in range(4):
            matrix.add_column(justify="center", ratio=1)

        snapshot_map = {snapshot.kind: snapshot for snapshot in status.snapshots}
        colors = self._snapshot_colors(status)
        status_cells: list[Text] = []
        detail_cells: list[Text] = []

        ordered = [
            (SNAPSHOT_KIND_CADDY_TUI, "caddy-tui"),
            (SNAPSHOT_KIND_CADDYFILE, "caddyfile"),
            (SNAPSHOT_KIND_CADDY_LIVE, "live caddy"),
        ]
        for kind, label in ordered:
            snapshot = snapshot_map.get(kind)
            status_cells.append(self._snapshot_status_cell(label, snapshot, colors.get(kind)))
            detail_cells.append(self._snapshot_detail_cell(snapshot))

        status_cells.append(self._service_status_cell(status))
        detail_cells.append(self._service_detail_cell(status))

        matrix.add_row(*status_cells)
        matrix.add_row(*detail_cells)
        return matrix

    @staticmethod
    def _snapshot_status_cell(label: str, snapshot: SnapshotInfo | None, color: str | None) -> Text:
        if snapshot and snapshot.available:
            style = color or "green"
            return Text(label, style=style)
        return Text(f"{label} (missing)", style=color or "grey50")

    @staticmethod
    def _snapshot_detail_cell(snapshot: SnapshotInfo | None) -> Text:
        if not snapshot or not snapshot.available:
            return Text("—", style="grey50")
        detail = f"{snapshot.block_count} block(s)"
        if snapshot.error:
            detail = f"{detail} • error: {snapshot.error}"
        return Text(detail)

    def _build_menu(self, status: AppStatus) -> list[MenuOption]:
        options: list[MenuOption] = [
            MenuOption("f", "Write Caddyfile ➜ caddy-tui", self._write_caddyfile_over_tui),
            MenuOption("t", "Write caddy-tui ➜ Caddyfile", self._write_tui_over_caddyfile),
            MenuOption("r", "Refresh live snapshot", self._refresh_live_snapshot),
            MenuOption("b", "Show snapshot blocks", self._show_snapshot_blocks),
        ]
        version_info = self._version_info
        if version_info and version_info.update_available and version_info.latest:
            options.append(MenuOption("u", f"Update available → {version_info.latest}", self._show_update_instructions))
        options.append(MenuOption("n", "Add caddy-tui block", self._add_caddy_tui_block))
        options.append(MenuOption("e", "Edit caddy-tui block", self._edit_caddy_tui_block))
        options.append(MenuOption("x", "Delete caddy-tui block", self._delete_caddy_tui_block))
        service_state = status.service_status.state if status.service_status else None
        if service_state == "down":
            options.append(MenuOption("s", "Restart Caddy (sudo)", self._restart_caddy))
        else:
            options.append(MenuOption("c", "Reload Caddy (sudo)", self._reload_caddy))
        if status.last_import_path:
            options.append(
                MenuOption(
                    "d",
                    "Show drift diff",
                    lambda path=status.last_import_path, db_path=status.db_path: self._show_drift(Path(path), db_path),
                )
            )
        options.append(MenuOption("q", "Quit", self._quit))
        return options

    # Actions

    def _write_caddyfile_over_tui(self) -> None:
        explicit_path: Path | None = None
        while True:
            try:
                summary = import_caddyfile(explicit_path, helper_interactive=True)
            except FileNotFoundError:
                explicit_path = self._prompt_for_path("Path to Caddyfile")
                if explicit_path is None:
                    self._set_message("Write cancelled.", style="yellow")
                    return
                continue
            except CaddyfilePermissionError as exc:
                self._set_message(str(exc), style="yellow")
                return
            except Exception as exc:  # pragma: no cover - unexpected failures bubble to user
                self._set_message(f"Failed to load Caddyfile: {exc}", style="red")
                return
            else:
                self._set_message(
                    f"Wrote Caddyfile ➜ caddy-tui ({summary.site_count} block(s))",
                    style="green",
                )
                self._schedule_live_refresh()
                return

    def _write_tui_over_caddyfile(self) -> None:
        status = self._latest_status
        default_path = Path(status.last_import_path) if status and status.last_import_path else None
        target = self._prompt_for_path("Overwrite which Caddyfile?", default=default_path)
        if target is None:
            self._set_message("Write cancelled.", style="yellow")
            return
        try:
            generate_caddyfile(target)
        except PermissionError as exc:
            self._set_message(str(exc), style="red")
            return
        except Exception as exc:  # pragma: no cover - unexpected failures bubble to user
            self._set_message(f"Failed to write Caddyfile: {exc}", style="red")
            return
        self._set_message(f"Wrote caddy-tui ➜ {target}", style="green")
        self._schedule_live_refresh()

    def _prompt_for_path(self, prompt: str, default: Path | None = None) -> Path | None:
        label = prompt
        if default:
            label += f" [{default}]"
        label += ": "
        response = input(label).strip()
        if not response:
            return default
        return Path(response).expanduser()

    def _show_drift(self, target: Path, db_path: Path) -> None:
        report = compare_caddyfile(target, db_path=db_path)
        if report.error:
            self._set_message(f"Unable to compare drift: {report.error}", style="red")
            return
        summary = summarise_drift(report)
        diff_text = report.diff or "No differences detected."
        border = "red" if report.diff else "green"
        panel = Panel(diff_text, title=summary, border_style=border, expand=False)
        self._set_renderable(panel)

    def _quit(self) -> None:
        self.running = False
        self._set_message("Goodbye!", style="green")

    def _show_update_instructions(self) -> None:
        info = self._get_version_info(force=True)
        if info is None:
            self._set_message("Unable to contact GitHub for version info. Try again later.", style="red")
            return
        if not info.update_available or not info.latest:
            self._set_message(f"Already running the latest version ({info.current}).", style="green")
            return
        install_cmd = "pip install --upgrade caddy-tui"
        panel = Panel(
            f"Current version: {info.current}\nLatest release: {info.latest}\n\nUpgrade with:\n  {install_cmd}",
            title="Update available",
            border_style="yellow",
        )
        self._set_renderable(panel)

    # Output helpers

    def _set_message(self, message: str, *, style: str | None = None, persist: bool = True) -> None:
        text = Text(message, style=style)
        self.last_output = text
        self._last_output_once = not persist
        if persist:
            self.console.print()
            self.console.print(text)

    def _set_renderable(self, renderable: RenderableType, *, persist: bool = True) -> None:
        self.last_output = renderable
        self._last_output_once = not persist
        if persist:
            self.console.print()
            self.console.print(renderable)

    @staticmethod
    def _format_comparison(comparison: SnapshotComparison) -> str:
        if comparison.status == "match":
            return "[green]match[/green]"
        if comparison.status == "different":
            mismatch = comparison.mismatch_count or 0
            suffix = f"{mismatch} block{'s' if mismatch != 1 else ''} differ"
            return f"[red]different[/red] ({suffix})"
        return "[yellow]missing[/yellow]"

    def _version_status_text(self) -> str:
        info = self._version_info
        if info is None:
            return "checking…"
        if info.update_available and info.latest:
            return f"[red]{info.current}[/red] → [yellow]{info.latest}[/yellow] (press 'u')"
        return f"[green]{info.current}[/green]"

    def _service_status_cell(self, status: AppStatus) -> Text:
        service = status.service_status
        if service is None:
            return Text("caddy service (n/a)", style="grey50")
        label = f"caddy service ({service.state})"
        return Text(label, style=self._service_color(status, service))

    def _service_detail_cell(self, status: AppStatus) -> Text:
        service = status.service_status
        if service is None:
            return Text("configure CADDY_TUI_ADMIN_ENDPOINT", style="grey50")
        parts: list[str] = []
        if service.block_count is not None:
            parts.append(f"{service.block_count} block(s)")
        if service.detail and service.detail.lower() not in {"live", "down", "unknown"}:
            parts.append(service.detail)
        if service.error:
            parts.append(f"error: {service.error}")
        if not parts:
            parts.append("no data")
        return Text(" • ".join(parts), style=self._service_color(status, service))

    def _snapshot_colors(self, status: AppStatus) -> dict[str, str | None]:
        snapshot_map = {snapshot.kind: snapshot for snapshot in status.snapshots}
        colors: dict[str, str | None] = {}
        for kind in (
            SNAPSHOT_KIND_CADDY_TUI,
            SNAPSHOT_KIND_CADDYFILE,
            SNAPSHOT_KIND_CADDY_LIVE,
        ):
            snap = snapshot_map.get(kind)
            if not snap or not snap.available:
                colors[kind] = "grey50"
            else:
                colors[kind] = None

        def _comparison_status(left: str, right: str) -> str | None:
            key = frozenset((left, right))
            for comparison in status.comparisons:
                if frozenset((comparison.left_kind, comparison.right_kind)) == key:
                    return comparison.status
            return None

        ct_cf = _comparison_status(SNAPSHOT_KIND_CADDY_TUI, SNAPSHOT_KIND_CADDYFILE)
        ct_live = _comparison_status(SNAPSHOT_KIND_CADDY_TUI, SNAPSHOT_KIND_CADDY_LIVE)

        tui_snap = snapshot_map.get(SNAPSHOT_KIND_CADDY_TUI)
        caddyfile_snap = snapshot_map.get(SNAPSHOT_KIND_CADDYFILE)
        live_snap = snapshot_map.get(SNAPSHOT_KIND_CADDY_LIVE)

        if tui_snap and tui_snap.available and caddyfile_snap and caddyfile_snap.available:
            if ct_cf == "match":
                colors[SNAPSHOT_KIND_CADDY_TUI] = "green"
                colors[SNAPSHOT_KIND_CADDYFILE] = "green"
            elif ct_cf == "different":
                colors[SNAPSHOT_KIND_CADDY_TUI] = "red"
                colors[SNAPSHOT_KIND_CADDYFILE] = "red"
            else:
                colors[SNAPSHOT_KIND_CADDY_TUI] = colors[SNAPSHOT_KIND_CADDYFILE] = "yellow"

        if not live_snap or not live_snap.available:
            colors[SNAPSHOT_KIND_CADDY_LIVE] = "grey50"
            return colors

        if ct_cf == "different":
            colors[SNAPSHOT_KIND_CADDY_LIVE] = "orange1"
        else:
            if ct_live == "match":
                colors[SNAPSHOT_KIND_CADDY_LIVE] = "green"
            elif ct_live == "different":
                colors[SNAPSHOT_KIND_CADDY_LIVE] = "red"
            else:
                colors[SNAPSHOT_KIND_CADDY_LIVE] = "yellow"

        return colors

    @staticmethod
    def _has_disagreement(status: AppStatus) -> bool:
        return any(comparison.status == "different" for comparison in status.comparisons)

    def _service_color(self, status: AppStatus, service: "ServiceStatus") -> str:
        if service.state == "down":
            return "red"
        if service.state != "live":
            return "yellow"
        return "green" if not self._has_disagreement(status) else "orange1"

    def _get_version_info(self, *, force: bool = False) -> VersionInfo | None:
        if self._version_info is not None and not force:
            return self._version_info
        now = time.monotonic()
        if not force and self._version_last_attempt and (now - self._version_last_attempt) < 300:
            return self._version_info
        return self._refresh_version_info()

    def _refresh_version_info(self) -> VersionInfo | None:
        self._version_last_attempt = time.monotonic()
        try:
            info = collect_version_info()
        except Exception:
            return None
        self._version_info = info
        return info

    # Live refresh helpers

    def _schedule_live_refresh(self) -> None:
        self._refresh_live_next = True

    def _refresh_live_snapshot(self) -> None:
        info = refresh_live_snapshot(live_caddyfile=LIVE_CADDYFILE)
        live_snapshot = next((snap for snap in info.snapshots if snap.kind == SNAPSHOT_KIND_CADDY_LIVE), None)
        if live_snapshot and live_snapshot.available:
            details = f"{live_snapshot.block_count} block(s)"
            if live_snapshot.collected_at:
                details = f"{details} @ {live_snapshot.collected_at}"
            self._set_message(f"Live snapshot refreshed ({details}).", style="green")
        elif live_snapshot and live_snapshot.error:
            self._set_message(f"Live snapshot refresh failed: {live_snapshot.error}", style="red")
        else:
            self._set_message("Live snapshot refresh failed: no data returned.", style="red")

    def _reload_caddy(self) -> None:
        success, command, error = reload_caddy_service(None)
        if not success:
            hint = error or command or "reload failed"
            self._set_message(f"Reload failed: {hint}", style="red")
            return
        printable = command or "helper reload"
        self._set_message(f"Reload requested via {printable}.", style="green")
        self._schedule_live_refresh()

    def _restart_caddy(self) -> None:
        success, command, error = restart_caddy_service(None)
        if not success:
            hint = error or command or "restart failed"
            self._set_message(f"Restart failed: {hint}", style="red")
            return
        printable = command or "helper restart"
        self._set_message(f"Restart requested via {printable}.", style="green")
        self._schedule_live_refresh()

    def _show_snapshot_blocks(self) -> None:
        status = self._latest_status
        if status is None or not status.db_ready:
            self._set_message("No snapshot data available. Run an import first.", style="yellow")
            return

        kinds = [
            (SNAPSHOT_KIND_CADDY_TUI, "caddy-tui"),
            (SNAPSHOT_KIND_CADDYFILE, "caddyfile"),
            (SNAPSHOT_KIND_CADDY_LIVE, "live caddy"),
        ]

        blocks_by_kind: dict[str, list[SnapshotBlockText]] = {}
        block_maps: dict[str, dict[int, SnapshotBlockText]] = {}
        for kind, _label in kinds:
            blocks = sorted(load_snapshot_block_texts(status.db_path, kind), key=lambda b: b.block_index)
            blocks_by_kind[kind] = blocks
            block_maps[kind] = {block.block_index: block for block in blocks}

        if not any(blocks_by_kind.values()):
            self._set_message("No snapshot blocks to display.", style="yellow")
            return

        table = Table(title="Snapshot Blocks", show_lines=True, expand=True)
        table.add_column("Key", style="bold cyan", justify="center", width=20)
        for _kind, label in kinds:
            table.add_column(label, overflow="fold", no_wrap=False, justify="left")

        live_blocks = blocks_by_kind.get(SNAPSHOT_KIND_CADDY_LIVE, [])

        if live_blocks:
            other_kinds = [kind for kind, _label in kinds if kind != SNAPSHOT_KIND_CADDY_LIVE]
            match_results: dict[str, dict[int, SnapshotBlockText]] = {}
            leftovers: dict[str, list[SnapshotBlockText]] = {}
            for kind in other_kinds:
                matches, remaining = self._match_blocks_by_tokens(live_blocks, blocks_by_kind.get(kind, []))
                match_results[kind] = matches
                leftovers[kind] = remaining

            for live_block in live_blocks:
                key_label = self._block_label(live_block)
                cells = [Text(key_label, style="bold cyan")]
                for kind, _label in kinds:
                    if kind == SNAPSHOT_KIND_CADDY_LIVE:
                        cells.append(self._format_block_cell(live_block))
                        continue
                    match = match_results.get(kind, {}).get(live_block.block_index)
                    if match:
                        cells.append(self._format_block_cell(match))
                    else:
                        cells.append(Text("—", style="grey50"))
                table.add_row(*cells)

            for kind in other_kinds:
                for block in leftovers.get(kind, []):
                    label = f"{SNAPSHOT_LABELS[kind]}(?)"
                    cells = [Text(label, style="bold cyan")]
                    for loop_kind, _label in kinds:
                        if loop_kind == kind:
                            cells.append(self._format_block_cell(block))
                        else:
                            cells.append(Text("—", style="grey50"))
                    table.add_row(*cells)
        else:
            block_indexes: set[int] = set()
            for block_map in block_maps.values():
                block_indexes.update(block_map.keys())

            for block_index in sorted(block_indexes):
                representative = next((bm.get(block_index) for bm in block_maps.values() if bm.get(block_index)), None)
                key_label = self._block_label(representative)
                cells = [Text(key_label, style="bold cyan")]
                for kind, _label in kinds:
                    entry = block_maps.get(kind, {}).get(block_index)
                    cells.append(self._format_block_cell(entry))
                table.add_row(*cells)

        self._set_renderable(table, persist=False)

    def _add_caddy_tui_block(self) -> None:
        result = self._load_blocks_for_editor(require_existing=False)
        if result is None:
            return
        db_path, blocks = result
        template = "example.com {\n    respond \"hello\"\n}\n"
        content = self._launch_editor(template)
        if content is None:
            return
        try:
            new_block = parse_single_block(content)
        except Exception as exc:
            self._set_message(f"Invalid block: {exc}", style="red")
            return
        blocks.append(new_block)
        try:
            save_caddy_tui_blocks(blocks, db_path)
        except Exception as exc:
            blocks.pop()
            self._set_message(f"Failed to save block: {exc}", style="red")
            return
        self._set_message(f"Added block #{len(blocks)} to caddy-tui snapshot.", style="green")

    def _edit_caddy_tui_block(self) -> None:
        result = self._load_blocks_for_editor(require_existing=True)
        if result is None:
            return
        db_path, blocks = result
        self._display_caddy_tui_blocks(blocks, title="Edit block")
        selection = self._prompt_block_selection(len(blocks), "edit")
        if selection is None:
            return
        current = blocks[selection]
        content = blocks_to_text([current]) or ""
        edited = self._launch_editor(content)
        if edited is None:
            return
        if edited == content:
            self._set_message("No changes made.", style="yellow")
            return
        try:
            replacement = parse_single_block(edited)
        except Exception as exc:
            self._set_message(f"Invalid block: {exc}", style="red")
            return
        original = blocks[selection]
        blocks[selection] = replacement
        try:
            save_caddy_tui_blocks(blocks, db_path)
        except Exception as exc:
            blocks[selection] = original
            self._set_message(f"Failed to save edits: {exc}", style="red")
            return
        self._set_message(f"Updated block #{selection + 1}.", style="green")

    def _delete_caddy_tui_block(self) -> None:
        result = self._load_blocks_for_editor(require_existing=True)
        if result is None:
            return
        db_path, blocks = result
        self._display_caddy_tui_blocks(blocks, title="Delete block")
        selection = self._prompt_block_selection(len(blocks), "delete")
        if selection is None:
            return
        confirm = input(f"Delete block #{selection + 1}? (y/N): ").strip().lower()
        if confirm not in {"y", "yes"}:
            self._set_message("Delete cancelled.", style="yellow")
            return
        removed = blocks.pop(selection)
        try:
            save_caddy_tui_blocks(blocks, db_path)
        except Exception as exc:
            blocks.insert(selection, removed)
            self._set_message(f"Failed to delete block: {exc}", style="red")
            return
        self._set_message(f"Deleted block #{selection + 1}.", style="green")

    def _load_blocks_for_editor(self, *, require_existing: bool) -> tuple[Path, list[ParsedBlock]] | None:
        status = self._latest_status
        if status is None or not status.db_ready:
            self._set_message("Database is not initialised. Run an import first.", style="yellow")
            return None
        db_path = status.db_path
        blocks = load_caddy_tui_blocks(db_path)
        if require_existing and not blocks:
            self._set_message("No caddy-tui snapshot available. Run an import first.", style="yellow")
            return None
        return db_path, list(blocks)

    def _display_caddy_tui_blocks(self, blocks: list[ParsedBlock], *, title: str) -> None:
        if not blocks:
            table = Table(title=title)
            table.add_column("#")
            table.add_column("Labels")
            table.add_column("Preview")
            table.add_row("—", "—", "No blocks available")
            self._set_renderable(table)
            return
        table = Table(title=title, show_lines=False)
        table.add_column("#", style="bold cyan", width=4, justify="right")
        table.add_column("Labels", style="bold", no_wrap=True)
        table.add_column("Preview", overflow="fold")
        for idx, block in enumerate(blocks, start=1):
            label = ", ".join(block.labels) if block.labels else "(global)"
            preview = self._parsed_block_preview(block)
            table.add_row(str(idx), label, preview)
        self._set_renderable(table)

    def _parsed_block_preview(self, block: ParsedBlock, *, max_length: int = 80) -> str:
        text = blocks_to_text([block]).strip()
        for line in text.splitlines():
            candidate = line.strip()
            if not candidate or candidate.startswith("#"):
                continue
            snippet = candidate
            break
        else:
            snippet = "(empty)"
        if len(snippet) > max_length:
            return snippet[: max_length - 1] + "…"
        return snippet

    def _prompt_block_selection(self, total: int, action: str) -> int | None:
        if total <= 0:
            self._set_message("No blocks available.", style="yellow")
            return None
        prompt = f"Select block number to {action} [1-{total}] (blank to cancel): "
        while True:
            choice = input(prompt).strip()
            if not choice:
                self._set_message(f"{action.capitalize()} cancelled.", style="yellow")
                return None
            if not choice.isdigit():
                self._set_message("Enter a numeric block index.", style="yellow")
                continue
            index = int(choice)
            if not 1 <= index <= total:
                self._set_message("Block number out of range.", style="yellow")
                continue
            return index - 1

    def _launch_editor(self, initial_text: str) -> str | None:
        command = self._resolve_editor_command()
        if command is None:
            return None
        initial = initial_text if initial_text.endswith("\n") else f"{initial_text}\n"
        with tempfile.NamedTemporaryFile("w+", delete=False, encoding="utf-8") as handle:
            handle.write(initial)
            handle.flush()
            temp_path = handle.name
        try:
            try:
                result = subprocess.run([*command, temp_path])
            except FileNotFoundError:
                self._set_message(f"Editor command not found: {command[0]}", style="red")
                return None
            except Exception as exc:
                self._set_message(f"Failed to launch editor: {exc}", style="red")
                return None
            if result.returncode != 0:
                self._set_message(f"Editor exited with status {result.returncode}.", style="red")
                return None
            with open(temp_path, "r", encoding="utf-8") as handle:
                return handle.read()
        finally:
            try:
                os.unlink(temp_path)
            except OSError:
                pass

    def _resolve_editor_command(self) -> list[str] | None:
        candidates: list[str] = []
        env_editor = os.environ.get("EDITOR")
        if env_editor:
            candidates.append(env_editor)
        candidates.extend(["nano", "vi", "vim"])
        for candidate in candidates:
            if not candidate:
                continue
            try:
                parts = shlex.split(candidate)
            except ValueError:
                continue
            executable = parts[0]
            if os.path.isabs(executable) or "/" in executable or shutil.which(executable):
                return parts
        self._set_message("Set $EDITOR or install nano/vi for editing.", style="red")
        return None

    @staticmethod
    def _block_label(block: SnapshotBlockText | None) -> str:
        if block is None:
            return "—"
        if block.key:
            return ", ".join(block.key)
        return str(block.block_index + 1)

    def _format_block_cell(
        self,
        block: SnapshotBlockText | None,
    ) -> Text:
        if block is None:
            return Text("—", style="grey50")
        display = block.text or "(empty)"
        return Text(display)

    @classmethod
    def _match_blocks_by_tokens(
        cls,
        source_blocks: list[SnapshotBlockText],
        target_blocks: list[SnapshotBlockText],
    ) -> tuple[dict[int, SnapshotBlockText], list[SnapshotBlockText]]:
        if not source_blocks or not target_blocks:
            return {}, target_blocks

        target_entries = [
            (block, cls._block_search_blob(block))
            for block in target_blocks
        ]
        available = list(target_entries)
        matches: dict[int, SnapshotBlockText] = {}

        for source in source_blocks:
            source_tokens = cls._block_tokens(source)
            if not source_tokens:
                continue
            best_idx = None
            best_score = 0.0
            for idx, (candidate, candidate_blob) in enumerate(available):
                score = cls._token_overlap_score(source_tokens, candidate_blob)
                if score > best_score:
                    best_score = score
                    best_idx = idx
            if best_idx is not None and best_score > 0:
                match_block, _blob = available.pop(best_idx)
                matches[source.block_index] = match_block

        leftovers = [entry[0] for entry in available]
        return matches, leftovers

    @staticmethod
    def _block_tokens(block: SnapshotBlockText | None) -> set[str]:
        tokens: set[str] = set()
        if block is None:
            return tokens

        def _extend(values: Iterable[str]) -> None:
            for value in values:
                trimmed = value.strip().lower()
                if trimmed:
                    tokens.add(trimmed)

        _extend(block.hosts)
        _extend(block.paths)
        _extend(block.groups)
        _extend(block.roots)
        _extend(block.dials)
        _extend(block.locations)
        _extend(block.encodings)
        _extend(block.status_codes)
        _extend(block.handlers)
        _extend(block.key)
        return tokens

    @staticmethod
    def _block_search_blob(block: SnapshotBlockText | None) -> str:
        if block is None:
            return ""

        segments: list[str] = []

        def _extend(values: Iterable[str]) -> None:
            for value in values:
                trimmed = value.strip().lower()
                if trimmed:
                    segments.append(trimmed)

        _extend(block.hosts)
        _extend(block.paths)
        _extend(block.groups)
        _extend(block.roots)
        _extend(block.dials)
        _extend(block.locations)
        _extend(block.encodings)
        _extend(block.status_codes)
        _extend(block.handlers)
        _extend(block.key)
        if block.text:
            segments.append(block.text.lower())
        return "\n".join(segments)

    @staticmethod
    def _token_overlap_score(source_tokens: list[str], candidate_blob: str) -> float:
        if not source_tokens:
            return 0.0
        if not candidate_blob:
            return 0.0
        matches = 0
        for token in source_tokens:
            if token and token in candidate_blob:
                matches += 1
        return matches / len(source_tokens)


def run_tui() -> None:
    TerminalMenuApp().run()



from pathlib import Path

from sqlalchemy import select

from caddy_tui import db
from caddy_tui.importer import import_caddyfile
from caddy_tui.models import (
    Config,
    ConfigSnapshot,
    RawFragment,
    ServerBlock,
    ServerBlockSite,
    SNAPSHOT_KIND_CADDYFILE,
    SNAPSHOT_KIND_CADDY_TUI,
)
from caddy_tui.exporter import render_caddyfile_text
from caddy_tui.drift import compare_caddyfile, summarise_drift
from caddy_tui import status as status_mod


def _reset_db(tmp_path: Path, sample: str | None = None) -> Path:
    db._engine = None  # type: ignore[attr-defined]
    db._SessionLocal = None  # type: ignore[attr-defined]
    db_path = tmp_path / "config.db"
    db.init_db(db_path)
    block_text = sample or "example.com {\n    respond \"ok\"\n}\n"
    with db.session_scope(db_path) as session:
        config = Config(name="default", caddyfile_path="/etc/caddy/Caddyfile")
        session.add(config)
        session.flush()
        snapshot = ConfigSnapshot(config=config, source_kind=SNAPSHOT_KIND_CADDY_TUI)
        session.add(snapshot)
        session.flush()
        block = ServerBlock(snapshot=snapshot, block_index=0, raw_prelude=None, raw_postlude=None)
        session.add(block)
        session.flush()
        block.sites.append(
            ServerBlockSite(
                raw_label="example.com",
                host="example.com",
                port=None,
                scheme=None,
                is_ipv6=False,
                is_wildcard=False,
                label_index=0,
            )
        )
        block.fragments.append(RawFragment(fragment_index=0, kind="block", content=block_text))
    return db_path


def test_compare_caddyfile_in_sync(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    target = tmp_path / "Caddyfile"
    target.write_text(render_caddyfile_text(db_path=db_path))

    report = compare_caddyfile(target, db_path=db_path)
    assert report.in_sync is True
    assert report.error is None
    assert report.diff is None


def test_compare_caddyfile_detects_drift(tmp_path: Path):
    db_path = _reset_db(tmp_path)
    target = tmp_path / "Caddyfile"
    target.write_text("broken")

    report = compare_caddyfile(target, db_path=db_path)
    assert report.in_sync is False
    assert report.diff and "broken" in report.diff
    assert report.error is None


def test_summarise_drift_respects_error(tmp_path: Path, monkeypatch):
    db_path = _reset_db(tmp_path)
    target = tmp_path / "Caddyfile"

    def _denied(*_args, **_kwargs):
        raise PermissionError

    monkeypatch.setattr(Path, "read_text", _denied)
    monkeypatch.setattr(
        "caddy_tui.drift.stage_caddyfile_copy",
        lambda path: (None, "sudo helper", "Permission denied"),
    )
    report = compare_caddyfile(target, db_path=db_path)
    assert "Permission denied" in summarise_drift(report)


def test_compare_caddyfile_uses_helper_on_permission(monkeypatch, tmp_path: Path):
    db_path = _reset_db(tmp_path)
    target = tmp_path / "Caddyfile"
    target.write_text("ignored")

    mirror = tmp_path / "mirror"
    mirror.write_text(render_caddyfile_text(db_path=db_path))

    original_read = Path.read_text

    def fake_read(self, *args, **kwargs):
        if self == target:
            raise PermissionError
        return original_read(self, *args, **kwargs)

    monkeypatch.setattr(Path, "read_text", fake_read)
    monkeypatch.setattr(
        "caddy_tui.drift.stage_caddyfile_copy",
        lambda path: (mirror, "sudo helper", None),
    )

    report = compare_caddyfile(target, db_path=db_path)
    assert report.in_sync is True
    assert report.error is None


def test_collect_app_status_includes_drift(tmp_path: Path, monkeypatch):
    db_path = _reset_db(tmp_path)
    target = tmp_path / "Caddyfile"
    target.write_text(render_caddyfile_text(db_path=db_path))
    with db.session_scope(db_path) as session:
        config = session.scalar(select(Config))
        assert config is not None
        config.caddyfile_path = str(target)
        config.last_imported_at = "2025-11-22T00:00:00Z"

    monkeypatch.setattr("caddy_tui.importer.adapt_caddyfile", lambda path: {})
    import_caddyfile(target, target_snapshot=SNAPSHOT_KIND_CADDY_TUI)
    info = status_mod.collect_app_status(db_path=db_path, refresh_sources=False)
    assert info.block_count == 1
    assert any(snapshot.block_count == 1 for snapshot in info.snapshots)
    assert any(
        comparison.status == "match"
        and {comparison.left_kind, comparison.right_kind} == {SNAPSHOT_KIND_CADDY_TUI, SNAPSHOT_KIND_CADDYFILE}
        for comparison in info.comparisons
    )



"""Render live snapshot JSON routes into Caddyfile-style text."""
from __future__ import annotations

from typing import Any, Iterable
import json

from . import models


INDENT_SPACES = 4
REDIRECT_CODES = {301, 302, 303, 307, 308}


def render_live_block_like_caddyfile(block: models.ServerBlock) -> str | None:
    """Best-effort conversion of a live JSON route block to Caddyfile text."""
    fragment = _first_json_route_fragment(block)
    if fragment is None:
        return None
    try:
        route = json.loads(fragment.content)
    except json.JSONDecodeError:
        return None
    if not isinstance(route, dict):
        return None

    header = _block_header_line(block)
    body_lines = _render_route_body(route, indent=INDENT_SPACES)
    if not body_lines:
        body_lines = [_indent(INDENT_SPACES) + "# no handlers"]

    lines = [header, *body_lines, "}"]
    body_text = "\n".join(lines) + "\n"
    return f"{block.raw_prelude or ''}{body_text}{block.raw_postlude or ''}"


def _first_json_route_fragment(block: models.ServerBlock) -> models.RawFragment | None:
    for fragment in sorted(block.fragments, key=lambda f: f.fragment_index):
        if fragment.kind == "json_route":
            return fragment
    return None


def _block_header_line(block: models.ServerBlock) -> str:
    labels = [
        site.raw_label.strip()
        for site in sorted(block.sites, key=lambda s: s.label_index)
        if site.raw_label and site.raw_label.strip()
    ]
    if labels:
        return f"{', '.join(labels)} {{"
    return "{"


def _render_route_body(route: dict[str, Any], *, indent: int) -> list[str]:
    lines: list[str] = []
    lines.extend(_match_comment_lines(route, indent))
    if route.get("terminal"):
        lines.append(_indent(indent) + "# terminal")

    handles = _handle_entries(route)
    if not handles:
        nested_routes = route.get("routes")
        if isinstance(nested_routes, list):
            for nested in nested_routes:
                lines.append(_indent(indent) + "handle {")
                lines.extend(_render_route_body(nested, indent=indent + INDENT_SPACES))
                lines.append(_indent(indent) + "}")
        return lines

    for entry in handles:
        lines.extend(_render_handle_entry(entry, indent))
    return lines


def _match_comment_lines(route: dict[str, Any], indent: int) -> list[str]:
    matchers = route.get("match")
    if not isinstance(matchers, list):
        return []
    lines: list[str] = []
    for matcher in matchers:
        if not isinstance(matcher, dict):
            continue
        desc = _describe_matcher(matcher)
        if desc:
            lines.append(_indent(indent) + f"# match {desc}")
    return lines


def _describe_matcher(matcher: dict[str, Any]) -> str:
    parts: list[str] = []
    hosts = _string_list(matcher.get("host")) or _string_list(matcher.get("hosts"))
    if hosts:
        parts.append("host " + ", ".join(hosts))
    paths = _string_list(matcher.get("path")) + _string_list(matcher.get("paths"))
    if paths:
        parts.append("path " + ", ".join(paths))
    methods = _string_list(matcher.get("method")) + _string_list(matcher.get("methods"))
    if methods:
        parts.append("method " + ", ".join(methods))
    if matcher.get("expression"):
        parts.append(f"expr {matcher['expression']}")
    return "; ".join(parts)


def _handle_entries(node: dict[str, Any]) -> list[dict[str, Any]]:
    handle_value = node.get("handle")
    handlers_value = node.get("handlers")
    entries = handle_value if handle_value is not None else handlers_value
    if isinstance(entries, dict):
        return [entries]
    if isinstance(entries, list):
        return [entry for entry in entries if isinstance(entry, dict)]
    return []


def _render_handle_entry(entry: dict[str, Any], indent: int) -> list[str]:
    handler = (entry.get("handler") or "").lower()
    if handler == "subroute":
        return _render_subroute(entry, indent)
    if handler == "reverse_proxy":
        return [_indent(indent) + _render_reverse_proxy(entry)]
    if handler == "static_response":
        return _render_static_response(entry, indent)
    if handler == "encode":
        return [_indent(indent) + _render_encode(entry)]
    if handler == "file_server":
        return _render_file_server(entry, indent)
    if handler in {"headers", "header"}:
        return _render_header(entry, indent)
    if handler == "php_fastcgi":
        return _render_php_fastcgi(entry, indent)
    if handler == "handle_response":
        return _render_handle_response(entry, indent)
    if handler == "rewrite":
        return _render_rewrite(entry, indent)
    if handler == "copy_response_headers":
        return _render_copy_response_headers(entry, indent)
    if handler == "request_body":
        return _render_request_body(entry, indent)
    return [_indent(indent) + f"# handler {handler or 'unknown'}"]


def _render_subroute(entry: dict[str, Any], indent: int) -> list[str]:
    routes = entry.get("routes")
    if not isinstance(routes, list):
        return [_indent(indent) + "handle {}"]
    lines: list[str] = []
    for route in routes:
        if not isinstance(route, dict):
            continue
        lines.append(_indent(indent) + "handle {")
        nested_lines = _render_route_body(route, indent=indent + INDENT_SPACES)
        if not nested_lines:
            nested_lines = [_indent(indent + INDENT_SPACES) + "# no handlers"]
        lines.extend(nested_lines)
        lines.append(_indent(indent) + "}")
    if not lines:
        lines.append(_indent(indent) + "handle {}")
    return lines


def _render_reverse_proxy(entry: dict[str, Any]) -> str:
    targets = _reverse_proxy_targets(entry)
    line = "reverse_proxy"
    if targets:
        line += " " + " ".join(targets)
    return line


def _reverse_proxy_targets(entry: dict[str, Any]) -> list[str]:
    targets: list[str] = []
    upstreams = entry.get("upstreams")
    if isinstance(upstreams, list):
        for upstream in upstreams:
            if isinstance(upstream, dict):
                dial = upstream.get("dial")
                if isinstance(dial, str) and dial:
                    targets.append(dial)
    targets.extend(_string_list(entry.get("to")))
    return targets


def _render_static_response(entry: dict[str, Any], indent: int) -> list[str]:
    headers = entry.get("headers") or {}
    location = _first_header_value(headers, "Location")
    status_code = entry.get("status_code")
    body = entry.get("body") or entry.get("content")
    if location and isinstance(status_code, int) and status_code in REDIRECT_CODES and not body:
        return [_indent(indent) + f"redir {location} {status_code}"]
    pieces: list[str] = ["respond"]
    if body:
        pieces.append(_quote(str(body)))
    if status_code:
        pieces.append(str(status_code))
    return [_indent(indent) + " ".join(pieces)]


def _render_encode(entry: dict[str, Any]) -> str:
    value = entry.get("encodings") or entry.get("formats")
    names: list[str] = []
    if isinstance(value, dict):
        names.extend(str(key) for key in value.keys())
    else:
        names.extend(_string_list(value))
    line = "encode"
    if names:
        line += " " + " ".join(names)
    return line


def _render_file_server(entry: dict[str, Any], indent: int) -> list[str]:
    line = "file_server"
    if entry.get("browse"):
        line += " browse"
    lines = [_indent(indent) + line]
    root = entry.get("root")
    if isinstance(root, str) and root:
        lines.append(_indent(indent + INDENT_SPACES) + f"root {root}")
    index = entry.get("index")
    if isinstance(index, list):
        for value in index:
            lines.append(_indent(indent + INDENT_SPACES) + f"index {value}")
    elif isinstance(index, str) and index:
        lines.append(_indent(indent + INDENT_SPACES) + f"index {index}")
    return lines


def _render_header(entry: dict[str, Any], indent: int) -> list[str]:
    response = entry.get("response") or {}
    set_headers = response.get("set") or entry.get("set")
    lines: list[str] = []
    if isinstance(set_headers, dict):
        for key, values in set_headers.items():
            for value in _string_list(values):
                lines.append(_indent(indent) + f"header {key} {_quote(value)}")
    if not lines:
        lines.append(_indent(indent) + "header /* configure headers */")
    return lines


def _render_php_fastcgi(entry: dict[str, Any], indent: int) -> list[str]:
    upstream = entry.get("upstream") or entry.get("address")
    line = "php_fastcgi"
    if isinstance(upstream, str) and upstream:
        line += f" {upstream}"
    lines = [_indent(indent) + line]
    root = entry.get("root")
    if isinstance(root, str) and root:
        lines.append(_indent(indent + INDENT_SPACES) + f"root {root}")
    return lines


def _render_handle_response(entry: dict[str, Any], indent: int) -> list[str]:
    routes = entry.get("routes")
    if not isinstance(routes, list):
        return [_indent(indent) + "handle_response {}"]
    lines: list[str] = []
    for route in routes:
        if not isinstance(route, dict):
            continue
        lines.append(_indent(indent) + "handle_response {")
        lines.extend(_render_route_body(route, indent=indent + INDENT_SPACES))
        lines.append(_indent(indent) + "}")
    if not lines:
        lines.append(_indent(indent) + "handle_response {}")
    return lines


def _render_rewrite(entry: dict[str, Any], indent: int) -> list[str]:
    destination = entry.get("to") or entry.get("uri")
    if isinstance(destination, str) and destination:
        return [_indent(indent) + f"rewrite {destination}"]
    return [_indent(indent) + "rewrite"]


def _render_copy_response_headers(entry: dict[str, Any], indent: int) -> list[str]:
    headers = entry.get("headers")
    if isinstance(headers, list):
        joined = " ".join(headers)
        return [_indent(indent) + f"copy_response_headers {joined}"]
    return [_indent(indent) + "copy_response_headers"]


def _render_request_body(entry: dict[str, Any], indent: int) -> list[str]:
    if entry.get("action") == "replace" and isinstance(entry.get("value"), str):
        return [_indent(indent) + f"request_body replace {_quote(entry['value'])}"]
    return [_indent(indent) + "request_body"]


def _first_header_value(headers: dict[str, Any], key: str) -> str | None:
    candidates = headers.get(key) or headers.get(key.lower())
    if isinstance(candidates, str) and candidates:
        return candidates
    if isinstance(candidates, Iterable):
        for candidate in candidates:
            if isinstance(candidate, str) and candidate:
                return candidate
    return None


def _string_list(value: Any) -> list[str]:
    if value is None:
        return []
    if isinstance(value, str):
        return [value]
    if isinstance(value, Iterable):
        results: list[str] = []
        for entry in value:
            if isinstance(entry, str) and entry:
                results.append(entry)
        return results
    return []


def _quote(value: str) -> str:
    escaped = value.replace('"', '\\"')
    return f'"{escaped}"'


def _indent(width: int) -> str:
    return " " * width



"""Helpers for editing caddy-tui snapshot blocks."""
from __future__ import annotations

from datetime import datetime, timezone
from hashlib import sha256
from pathlib import Path
from typing import Iterable

from sqlalchemy import select

from . import models
from .caddyfile_parser import ParsedBlock, parse_caddyfile_text
from .db import session_scope
from .importer import DEFAULT_CONFIG_NAME, _write_snapshot
from .snapshots import get_snapshot, render_snapshot_text


EDITOR_SOURCE_LABEL = "caddy-tui editor"


def load_caddy_tui_blocks(db_path: Path) -> list[ParsedBlock]:
    """Return the parsed blocks for the caddy-tui snapshot."""
    text = _snapshot_text(db_path)
    if text is None:
        return []
    parsed = parse_caddyfile_text(text)
    return parsed.blocks


def save_caddy_tui_blocks(blocks: list[ParsedBlock], db_path: Path, *, source_label: str = EDITOR_SOURCE_LABEL) -> None:
    """Persist parsed blocks back to the caddy-tui snapshot."""
    text = blocks_to_text(blocks)
    parsed = parse_caddyfile_text(text)
    digest = sha256(text.encode("utf-8")).hexdigest()
    collected_at = datetime.now(timezone.utc)

    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
        if config is None:
            raise RuntimeError("Initialise the database before editing blocks.")
        _write_snapshot(
            session,
            config,
            models.SNAPSHOT_KIND_CADDY_TUI,
            parsed.blocks,
            source_path=source_label,
            source_hash=digest,
            collected_at=collected_at,
        )


def blocks_to_text(blocks: Iterable[ParsedBlock]) -> str:
    parts: list[str] = []
    for block in blocks:
        if block.raw_prelude:
            parts.append(block.raw_prelude)
        for fragment in block.fragments:
            parts.append(fragment.content)
        if block.raw_postlude:
            parts.append(block.raw_postlude)
    return "".join(parts)


def parse_single_block(text: str) -> ParsedBlock:
    """Parse ``text`` and ensure it contains exactly one server block."""
    parsed = parse_caddyfile_text(text)
    blocks = [block for block in parsed.blocks if block.fragments]
    if len(blocks) != 1:
        raise ValueError("Expected exactly one server block")
    return blocks[0]


def _snapshot_text(db_path: Path) -> str | None:
    with session_scope(db_path=db_path) as session:
        config = session.scalar(select(models.Config).where(models.Config.name == DEFAULT_CONFIG_NAME))
        if config is None:
            return None
        snapshot = get_snapshot(session, config.id, models.SNAPSHOT_KIND_CADDY_TUI)
        if snapshot is None:
            return None
        return render_snapshot_text(snapshot)



"""Version discovery and tracking utilities."""
from __future__ import annotations

import json
import os
import urllib.request
from dataclasses import dataclass
from typing import Optional

from packaging.version import InvalidVersion, Version

from . import __version__
from pathlib import Path

from .db import session_scope
from . import models

DEFAULT_REPO = "ajphilton/caddy-tui"


@dataclass(slots=True)
class VersionInfo:
    current: str
    latest: str | None
    update_available: bool
    source: str


def _normalize(version: str) -> Version | None:
    value = version.lstrip("v")
    try:
        return Version(value)
    except InvalidVersion:
        return None


def fetch_latest_version(repo: str | None = None, timeout: int = 5) -> str | None:
    """Query GitHub releases for the latest version tag."""
    repository = repo or os.environ.get("CADDY_TUI_REPO", DEFAULT_REPO)
    url = f"https://api.github.com/repos/{repository}/releases/latest"
    try:
        with urllib.request.urlopen(url, timeout=timeout) as resp:
            payload = json.loads(resp.read().decode("utf-8"))
    except Exception:
        return None
    tag = payload.get("tag_name") or payload.get("name")
    if not tag:
        return None
    normalized = _normalize(tag)
    return normalized.public if normalized else None


def collect_version_info(repo: str | None = None) -> VersionInfo:
    current = __version__
    latest = fetch_latest_version(repo)
    current_v = _normalize(current)
    latest_v = _normalize(latest) if latest else None
    update_available = bool(latest_v and current_v and latest_v > current_v)
    return VersionInfo(current=current, latest=latest, update_available=update_available, source=repo or os.environ.get("CADDY_TUI_REPO", DEFAULT_REPO))


def store_current_version(version: str | None = None, db_path: Path | str | None = None) -> None:
    value = version or __version__
    with session_scope(db_path=db_path) as session:
        try:
            meta = session.get(models.Meta, "app_version")
        except Exception:
            return
        if meta:
            meta.value = value
        else:
            session.add(models.Meta(key="app_version", value=value))


"""Application configuration helpers."""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import os

try:  # pragma: no cover - pwd isn't available on Windows
    import pwd
except ImportError:  # pragma: no cover
    pwd = None


def _determine_home() -> Path:
    sudo_user = os.environ.get("SUDO_USER")
    if sudo_user and pwd:
        try:
            return Path(pwd.getpwnam(sudo_user).pw_dir)
        except KeyError:  # pragma: no cover - only when user missing from passwd
            pass
    return Path.home()

APP_DIR = Path(os.environ.get("CADDY_TUI_HOME", _determine_home() / ".caddy-tui"))
DB_PATH = Path(os.environ.get("CADDY_TUI_DB", APP_DIR / "config.db"))
CACHE_DIR = Path(os.environ.get("CADDY_TUI_CACHE", APP_DIR / "cache"))
HELPER_SOCKET = Path(os.environ.get("CADDY_TUI_HELPER_SOCKET", APP_DIR / "helper.sock"))
GENERATED_CADDYFILE = Path(
    os.environ.get("CADDY_TUI_GENERATED_CADDYFILE", "/etc/caddy/Caddyfile.generated")
)
GENERATED_JSON = Path(
    os.environ.get("CADDY_TUI_GENERATED_JSON", "/etc/caddy/caddy-tui.json")
)
CADDY_BIN = os.environ.get("CADDY_TUI_CADDY_BIN")
RELOAD_MODE = os.environ.get("CADDY_TUI_RELOAD_MODE", "caddy")
LIVE_CADDYFILE = (
    Path(os.environ["CADDY_TUI_LIVE_CADDYFILE"]).expanduser()
    if os.environ.get("CADDY_TUI_LIVE_CADDYFILE")
    else None
)
CADDY_ADMIN_ENDPOINT = os.environ.get("CADDY_TUI_ADMIN_ENDPOINT", "http://127.0.0.1:2019/config")
CADDY_ADMIN_TIMEOUT = float(os.environ.get("CADDY_TUI_ADMIN_TIMEOUT", "2.5"))


@dataclass(slots=True)
class AppPaths:
    db_path: Path = DB_PATH
    generated_caddyfile: Path = GENERATED_CADDYFILE
    generated_json: Path = GENERATED_JSON
    caddy_bin: str | None = CADDY_BIN
    reload_mode: str = RELOAD_MODE
    live_caddyfile: Path | None = LIVE_CADDYFILE


def ensure_app_dir(path: Path | None = None) -> Path:
    """Ensure the data directory exists and return it."""
    target = path or APP_DIR
    target.mkdir(parents=True, exist_ok=True)
    return target


def ensure_cache_dir() -> Path:
    root = ensure_app_dir()
    cache = CACHE_DIR
    cache.mkdir(parents=True, exist_ok=True)
    return cache



"""Privileged helper commands for caddy-tui.

This module is intended to be executed via sudo. It exposes a very small surface
area so operators can grant password-less sudo access to `caddy-tui-helper`
without giving the TUI full root access.
"""
from __future__ import annotations

import os
import shutil
import stat
import subprocess
from pathlib import Path

import click


@click.group()
def main() -> None:
    """Run restricted privileged operations for caddy-tui."""


@main.command()
@click.option("--source", type=click.Path(exists=True, dir_okay=False, path_type=Path), required=True)
@click.option("--dest", type=click.Path(dir_okay=False, path_type=Path), required=True)
@click.option("--owner", type=int, required=True, help="Target file owner UID")
@click.option("--group", type=int, required=True, help="Target file group GID")
def mirror(source: Path, dest: Path, owner: int, group: int) -> None:
    """Copy a root-owned Caddyfile into an unprivileged staging area."""
    dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy2(source, dest)
    os.chown(dest, owner, group)
    click.echo(f"Mirrored {source} -> {dest}")


@main.command()
@click.option("--source", type=click.Path(exists=True, dir_okay=False, path_type=Path), required=True)
@click.option("--dest", type=click.Path(dir_okay=False, path_type=Path), required=True)
@click.option("--mode", type=str, default="0o644")
def install(source: Path, dest: Path, mode: str) -> None:
    """Install a generated Caddyfile into /etc with controlled permissions."""
    dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy2(source, dest)
    os.chmod(dest, int(mode, 8))
    click.echo(f"Installed {source} -> {dest}")


@main.command()
@click.option("--command", default="systemctl reload caddy", help="Reload command to execute.")
def reload(command: str) -> None:
    """Reload the running Caddy daemon."""
    parts = command.split()
    subprocess.run(parts, check=True)
    click.echo("Reloaded Caddy")


@main.command()
@click.option("--command", default="systemctl restart caddy", help="Restart command to execute.")
def restart(command: str) -> None:
    """Restart the Caddy daemon when it is not running."""
    parts = command.split()
    subprocess.run(parts, check=True)
    click.echo("Restarted Caddy")


@main.command(name="status")
@click.option("--command", default="systemctl is-active caddy", help="Command that reports Caddy service state.")
def status_cmd(command: str) -> None:
    """Report whether Caddy is live or down."""
    parts = command.split()
    proc = subprocess.run(parts, capture_output=True, text=True, check=False)
    output = (proc.stdout or "").strip()
    if proc.returncode != 0 and not output:
        message = (proc.stderr or "status command failed").strip()
        raise click.ClickException(message)
    click.echo(output or "unknown")


if __name__ == "__main__":
    main()



"""SQLAlchemy models for logical Caddyfile storage."""
from __future__ import annotations

from datetime import datetime, timezone
from typing import Any, Literal

from sqlalchemy import Boolean, DateTime, ForeignKey, Integer, String, Text, UniqueConstraint
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


class Base(DeclarativeBase):
    pass


SNAPSHOT_KIND_CADDY_TUI = "caddy_tui"
SNAPSHOT_KIND_CADDYFILE = "caddyfile"
SNAPSHOT_KIND_CADDY_LIVE = "caddy_live"
SnapshotKind = Literal[
    SNAPSHOT_KIND_CADDY_TUI,
    SNAPSHOT_KIND_CADDYFILE,
    SNAPSHOT_KIND_CADDY_LIVE,
]


class Config(Base):
    __tablename__ = "configs"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(64), unique=True, nullable=False)
    caddyfile_path: Mapped[str] = mapped_column(Text(), nullable=False)
    last_imported_at: Mapped[str | None] = mapped_column(String(64))
    last_exported_at: Mapped[str | None] = mapped_column(String(64))
    last_caddyfile_hash: Mapped[str | None] = mapped_column(String(128))

    snapshots: Mapped[list[ConfigSnapshot]] = relationship(
        back_populates="config",
        cascade="all, delete-orphan",
        order_by="ConfigSnapshot.id",
    )


class ConfigSnapshot(Base):
    __tablename__ = "config_snapshots"
    __table_args__ = (UniqueConstraint("config_id", "source_kind", name="uq_config_snapshot_kind"),)

    id: Mapped[int] = mapped_column(primary_key=True)
    config_id: Mapped[int] = mapped_column(ForeignKey("configs.id", ondelete="CASCADE"), index=True)
    source_kind: Mapped[str] = mapped_column(String(32), nullable=False)
    source_label: Mapped[str | None] = mapped_column(String(64))
    source_path: Mapped[str | None] = mapped_column(Text())
    source_hash: Mapped[str | None] = mapped_column(String(128))
    collected_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow)
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow)
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow, onupdate=_utcnow)

    config: Mapped[Config] = relationship(back_populates="snapshots")
    server_blocks: Mapped[list[ServerBlock]] = relationship(
        back_populates="snapshot",
        cascade="all, delete-orphan",
        order_by="ServerBlock.block_index",
    )


class ServerBlock(Base):
    __tablename__ = "server_blocks"

    id: Mapped[int] = mapped_column(primary_key=True)
    snapshot_id: Mapped[int] = mapped_column(ForeignKey("config_snapshots.id", ondelete="CASCADE"), index=True)
    block_index: Mapped[int] = mapped_column(Integer, nullable=False)
    is_global: Mapped[bool] = mapped_column(Boolean, default=False)
    raw_prelude: Mapped[str | None] = mapped_column(Text())
    raw_postlude: Mapped[str | None] = mapped_column(Text())
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow)
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow, onupdate=_utcnow)

    snapshot: Mapped[ConfigSnapshot] = relationship(back_populates="server_blocks")
    sites: Mapped[list[ServerBlockSite]] = relationship(
        back_populates="block",
        cascade="all, delete-orphan",
        order_by="ServerBlockSite.label_index",
    )
    directives: Mapped[list[Directive]] = relationship(
        back_populates="block",
        cascade="all, delete-orphan",
        order_by="Directive.line_index",
    )
    fragments: Mapped[list[RawFragment]] = relationship(
        back_populates="block",
        cascade="all, delete-orphan",
        order_by="RawFragment.fragment_index",
    )


class ServerBlockSite(Base):
    __tablename__ = "server_block_sites"

    id: Mapped[int] = mapped_column(primary_key=True)
    block_id: Mapped[int] = mapped_column(ForeignKey("server_blocks.id", ondelete="CASCADE"), index=True)
    raw_label: Mapped[str] = mapped_column(Text(), nullable=False)
    host: Mapped[str | None] = mapped_column(String(255))
    port: Mapped[int | None] = mapped_column(Integer)
    scheme: Mapped[str | None] = mapped_column(String(16))
    is_ipv6: Mapped[bool] = mapped_column(Boolean, default=False)
    is_wildcard: Mapped[bool] = mapped_column(Boolean, default=False)
    label_index: Mapped[int] = mapped_column(Integer, nullable=False)

    block: Mapped[ServerBlock] = relationship(back_populates="sites")


class Directive(Base):
    __tablename__ = "directives"

    id: Mapped[int] = mapped_column(primary_key=True)
    block_id: Mapped[int] = mapped_column(ForeignKey("server_blocks.id", ondelete="CASCADE"), index=True)
    name: Mapped[str] = mapped_column(String(128), nullable=False)
    matcher: Mapped[str | None] = mapped_column(String(128))
    line_index: Mapped[int] = mapped_column(Integer, nullable=False)
    raw_leading: Mapped[str | None] = mapped_column(Text())
    raw_trailing: Mapped[str | None] = mapped_column(Text())
    has_block: Mapped[bool] = mapped_column(Boolean, default=False)
    raw_block_body: Mapped[str | None] = mapped_column(Text())
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow)
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow, onupdate=_utcnow)

    block: Mapped[ServerBlock] = relationship(back_populates="directives")
    args: Mapped[list[DirectiveArg]] = relationship(
        back_populates="directive",
        cascade="all, delete-orphan",
        order_by="DirectiveArg.arg_index",
    )
    kv_pairs: Mapped[list[DirectiveKeyValue]] = relationship(
        back_populates="directive",
        cascade="all, delete-orphan",
        order_by="DirectiveKeyValue.kv_index",
    )


class DirectiveArg(Base):
    __tablename__ = "directive_args"

    id: Mapped[int] = mapped_column(primary_key=True)
    directive_id: Mapped[int] = mapped_column(ForeignKey("directives.id", ondelete="CASCADE"), index=True)
    arg_index: Mapped[int] = mapped_column(Integer, nullable=False)
    value: Mapped[str] = mapped_column(Text(), nullable=False)

    directive: Mapped[Directive] = relationship(back_populates="args")


class DirectiveKeyValue(Base):
    __tablename__ = "directive_kv"

    id: Mapped[int] = mapped_column(primary_key=True)
    directive_id: Mapped[int] = mapped_column(ForeignKey("directives.id", ondelete="CASCADE"), index=True)
    section: Mapped[str | None] = mapped_column(String(64))
    key: Mapped[str] = mapped_column(String(255), nullable=False)
    value: Mapped[str] = mapped_column(Text(), nullable=False)
    kv_index: Mapped[int] = mapped_column(Integer, nullable=False)

    directive: Mapped[Directive] = relationship(back_populates="kv_pairs")


class RawFragment(Base):
    __tablename__ = "raw_fragments"

    id: Mapped[int] = mapped_column(primary_key=True)
    block_id: Mapped[int] = mapped_column(ForeignKey("server_blocks.id", ondelete="CASCADE"), index=True)
    fragment_index: Mapped[int] = mapped_column(Integer, nullable=False)
    kind: Mapped[str] = mapped_column(String(64), nullable=False)
    content: Mapped[str] = mapped_column(Text(), nullable=False)

    block: Mapped[ServerBlock] = relationship(back_populates="fragments")


class Meta(Base):
    __tablename__ = "meta"

    key: Mapped[str] = mapped_column(String(64), primary_key=True)
    value: Mapped[str | None] = mapped_column(Text())
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), default=_utcnow, onupdate=_utcnow)


def to_dict(instance: Base) -> dict[str, Any]:
    """Return a dictionary of column values for debugging."""
    return {column.key: getattr(instance, column.key) for column in instance.__table__.columns}



"""Database bootstrap helpers."""
from __future__ import annotations

from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator

from sqlalchemy import create_engine, event, text
from sqlalchemy.orm import Session, sessionmaker

from . import models
from .config import DB_PATH, ensure_app_dir


_engine = None
_SessionLocal: sessionmaker[Session] | None = None


def get_engine(db_path: Path | str | None = None):
    """Return a singleton engine for the configured DB path."""
    global _engine, _SessionLocal
    if _engine is None:
        ensure_app_dir()
        path = Path(db_path or DB_PATH)
        _engine = create_engine(f"sqlite:///{path}", future=True)
        event.listen(_engine, "connect", _enable_foreign_keys)
        _SessionLocal = sessionmaker(bind=_engine, expire_on_commit=False, future=True)
        _bootstrap_schema(_engine)
    return _engine


def init_db(db_path: Path | str | None = None) -> None:
    """Create all tables."""
    engine = get_engine(db_path=db_path)
    models.Base.metadata.create_all(engine)


@contextmanager
def session_scope(db_path: Path | str | None = None) -> Iterator[Session]:
    """Provide a transactional scope."""
    if _SessionLocal is None:
        get_engine(db_path=db_path)
    assert _SessionLocal is not None  # safety
    session = _SessionLocal()
    try:
        yield session
        session.commit()
    except Exception:  # pragma: no cover
        session.rollback()
        raise
    finally:
        session.close()


def _bootstrap_schema(engine) -> None:
    """Create tables and apply lightweight migrations when needed."""
    models.Base.metadata.create_all(engine)
    _ensure_schema_version(engine)


def _enable_foreign_keys(dbapi_connection, _connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys = ON")
    cursor.close()


def _ensure_schema_version(engine) -> None:
    with engine.begin() as conn:
        columns = _table_columns(engine, "meta")
        if columns is None:
            return
        result = conn.execute(text("SELECT value FROM meta WHERE key = 'schema_version'"))
        if result.fetchone() is None:
            conn.execute(
                text(
                    "INSERT INTO meta (key, value, updated_at) "
                    "VALUES (:key, :value, :updated_at)"
                ),
                {
                    "key": "schema_version",
                    "value": "2",
                    "updated_at": datetime.now(timezone.utc),
                },
            )


def _table_columns(engine, table_name: str) -> set[str] | None:
    with engine.connect() as conn:
        result = conn.execute(
            text(
                "SELECT name FROM sqlite_master WHERE type='table' AND name = :name"
            ),
            {"name": table_name},
        )
        if result.fetchone() is None:
            return None
        columns = conn.execute(text(f"PRAGMA table_info({table_name})"))
        return {row[1] for row in columns}
